# net_contextdiff.__main__


# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>
#
# This script compares Cisco IOS configuration files and outputs a
# configuration file which will convert the former into the latter.



from net_contextdiff import __version__

import argparse
import re
import sys

from deepops import deepmerge
import yaml

from .ios import CiscoIOSConfig, CiscoIOSDiffConfig



# --- constants ---



# PLATFORMS = dict
#
# This dictionary specifies the available platforms available to compare
# configurations in.  The dictionary is keyed on the user-specified platform
# name and specifies two values in a tuple, giving the object classes to be
# used for that platform: the first is the parser an the second is the
# difference comparator.

PLATFORMS = {
    "ios": (CiscoIOSConfig, CiscoIOSDiffConfig)
}



# --- functions ---



def diffconfig(diff, devicename, from_filename, to_filename, output_filename,
               no_output, dump_rules, debug_parser, debug_convert):

    """This function compares the 'from' and 'to' configurations for
    the specified device and writes a difference configuration file
    (one that transforms the configuration of a running device from the
    'from' to the 'to' state).

    The filenames for the configuration files, as well as the output
    are taken from the arguments parsed above and stored in global
    variables used directly by this function.

    The function returns True iff the parsing and comparison succeeded.
    Most problems result in the function aborting with an exception,
    but some minor warnings may be ignored and the program continue
    with devices.

    Keyword arguments:

    diff -- the DiffConfig object to do the convert the configuration

    devicename -- the name of the device being converted, or None if
    it's not known

    from_filename -- filename of the source configuration

    to_filename -- filename of the target configuration

    output_filename -- name of the file to write the conversion
    configuration to or None for standard output

    no_output -- a boolean set to True iff no output file is to be
    written

    dump_rules, debug_parser, debug_convert -- the equivalent
    variables to command line arguments
    """


    # reset the rules list for this conversion - this is the same
    # list each time, except the devicename changes and may cause
    # some of the rules to change

    diff.init_rules_active()
    diff.add_rules_active(rule_specs, devicename)

    if dump_rules >= 2:
        num = 0
        for spec, tree in diff.get_rules():
            print(">> rule num: %d spec: %s" % (num, spec))
            print(yaml.dump(tree, default_flow_style=False)
                        if tree else "<none>\n", sep="\n")

            num += 1


    # read in the 'from' and 'to' configurations

    from_cfg = config_parser_class(filename=from_filename, debug=debug_parser)

    to_cfg = (config_parser_class(filename=to_filename, debug=debug_parser)
                  if to_filename else None)


    # find the differences

    diffs, diffs_tree = diff.convert(from_cfg, to_cfg)


    # if output is suppressed, just stop with success here

    if no_output:
        return diffs_tree


    # write the differences to a file or standard output

    if output_filename:
        if debug_convert:
            print("debug: writing to output file:", output_filename)

        with open(output_filename, "w") as output_file:
            if diffs:
                print(diffs, file=output_file)

    else:
        if debug_convert:
            print("debug: writing to standard output")

        if diffs:
            print(diffs)


    return diffs_tree



# --- command line arguments ---



# create the parser and add in the available command line options

parser = argparse.ArgumentParser(
    # override the program name as running this as a __main__ inside a module
    # directory will use '__main__' by default - this name isn't necessarily
    # correct, but it looks better than that
    prog="net-contextdiff",

    # we want the epilog help output to be printed as it and not reformatted or
    # line wrapped
    formatter_class=argparse.RawDescriptionHelpFormatter)

parser.add_argument(
    "-r", "--rules-tree-filename",
    metavar="FILENAME",
    help="read rules data from a YAML file (assumed to be an "
         "'excludes' list by default - see -s option)")

parser.add_argument(
    "-s", "--rule-spec",
    metavar="PATH",
    action="append",
    default=[],
    help="add a rule to the rules list; rules are in the form '!path' "
         "and processed in order: '!' excludes the specified items and "
         "omitting it includes ONLY the specified items; the 'path' is "
         "a dictionary path, with keys separated by colons; the path "
         "is into the file read with -r; default can be viewed with "
         "-R option)")

parser.add_argument(
    "-e", "--explain",
    action="store_true",
    help="explain which converter paths match a configuration change")

parser.add_argument(
    "-q", "--quiet",
    action="store_true",
    help="when generating configuration for multiple devices, don't "
         "print the name of each device, as it's generated")

parser.add_argument(
    "-O", "--no-output",
    action="store_true",
    help="do not write changes output file - used to test parsing, "
         "exclude and conversion process")

parser.add_argument(
    "-R", "--dump-rules",
    action="count",
    default=0,
    help="dump the tree of items to be included and excluded from the "
         "comparisons, based on the rules (multiple levels increase "
         "verbosity up to a maximum of 2)")

parser.add_argument(
    "-C", "--dump-config",
    action="store_true",
    help="dump the configuration dictionary after parsing")

parser.add_argument(
    "-D", "--dump-diff",
    action="store_true",
    help="dump the difference (removes and updates) between the from "
         "and to configurations")

parser.add_argument(
    "-T", "--dump-diff-tree",
    action="store_true",
    help="dump the differences across all compared files as a YAML "
         "tree - this is useful for using in an exclude file")

parser.add_argument(
    "-P", "--debug-parser",
    action="count",
    default=0,
    help="enable debugging of contextual parsing (multiple levels "
         "increase verbosity up to a maxmimum of 2)")

parser.add_argument(
    "-V", "--debug-convert",
    action="count",
    default=0,
    help="enable debugging of the difference converstion action "
         "processing (multiple levels increase verbosity up to a "
         "maximum of 2)")

parser.add_argument(
    "platform",
    choices=PLATFORMS,
    help="platform used for configuration files")

parser.add_argument(
    "from_filename",
    metavar="from",
    help="initial ('from') configuration file; '%%' can be used to "
         "substitute in the name of the device into the filename")

parser.add_argument(
    "to_filename",
    nargs="?",
    default=None,
    metavar="to",
    help="destination ('to') configuration file; '%%' can be used to "
         "substitute in the name of the device into the filename; if "
         "omitted just parse the file and assume 'debug' mode")

parser.add_argument(
    "output_filename",
    nargs="?",
    metavar="output",
    help="write differences configuration to named file instead of "
         "stdout; '%%' can be used to substitute in the name of the "
         "device into the filename")

parser.add_argument(
    "devicenames",
    metavar="devicename",
    nargs="*",
    help="name(s) of the device(s) to calculate differences in the "
         "configuration for")

parser.add_argument(
    "-v", "--version",
    action="version",
    version=("%(prog)s " + __version__))


# parse the supplied command line against these options, storing the results

args = parser.parse_args()

config_parser_class, config_diff_class = PLATFORMS[args.platform]

rules_tree_filename = args.rules_tree_filename
rule_specs = args.rule_spec
explain = args.explain
quiet = args.quiet
no_output = args.no_output
dump_rules = args.dump_rules
dump_config = args.dump_config
dump_diff = args.dump_diff
dump_diff_tree = args.dump_diff_tree
debug_parser = args.debug_parser
debug_convert = args.debug_convert
from_filename = args.from_filename
to_filename = args.to_filename
output_filename = args.output_filename
devicenames = args.devicenames


# if the rule specifications were not specified, assume this set by
# default (we can't add_argument(default=...) option because the
# 'action="append"' option just adds user-specified items to it)

if not rule_specs:
    rule_specs = [
        "!rules:exclude:ALL",
        "!rules:exclude:%",
        "rules:include:ALL",
        "rules:include:%",
    ]


# check a couple of nonsensical configurations aren't being use related
# to multiple devices

if not len(devicenames):
    if from_filename.find("%") != -1:
        print("error: no device names specified, so operating on a single "
              "file, yet 'from' filename has '%' character",
              file=sys.stderr)

        exit(1)


    if to_filename and (to_filename.find("%") != -1):
        print("error: no device names specified, so operating on a single "
              "file, yet 'to' filename has '%' character",
              file=sys.stderr)

        exit(1)


    if output_filename and (output_filename.find("%") != -1):
        print("error: no device names specified, so operating on a single "
              "file, yet 'output' filename has '%' character",
              file=sys.stderr)

        exit(1)


elif len(devicenames) > 1:
    if from_filename.find("%") == -1:
        print("warning: multiple device names specified but 'from' filename "
              "does not contain '%' - same file will be read",
              file=sys.stderr)

    if to_filename and (to_filename.find("%") == -1):
        print("warning: multiple device names specified but 'to' filename "
              "does not contain '%' - same file will be read",
              file=sys.stderr)


    if not output_filename:
        print("warning: multiple device names specified but outputting to "
              "standard output - all configurations will be concatenated",
              file=sys.stderr)

    elif output_filename.find("%") == -1:
        print("error: multiple device names specified but 'output' filename "
              "does not contain '%' - same file would be overwritten",
              file=sys.stderr)

        exit(1)



# --- setup ---



# create the diff object once at the start

diff = config_diff_class(explain, dump_config, dump_diff, debug_convert)


# if a rules tree file was specified, read that in

if rules_tree_filename:
    diff.add_rules_tree_file(rules_tree_filename)


# dump the rule tree, if debugging enabled for it (this will include any
# rules included by default, in the class)

if dump_rules >= 1:
    print(">> rule tree:")
    print(yaml.dump(diff.get_rules_tree(), default_flow_style=False))
    print(">> rule specs:")
    for spec in rule_specs:
        print("->", spec)

    print()



# --- compare ---



# if we have a list of device names (even just one), we iterate through
# them, else we just process the one file we've been given

if devicenames:
    # initialise the tree of aggregated differences across al the
    # compared configuration files

    all_diffs_tree = {}


    for devicename in devicenames:
        if not quiet:
            print(devicename)


        # do the actual conversion

        diffs_tree = diffconfig(
                      diff, devicename,
                      from_filename.replace("%", devicename),
                      to_filename.replace("%", devicename),
                      output_filename.replace("%", devicename),
                      no_output, dump_rules, debug_parser,
                      debug_convert)


        # add the differences in this conversion to the aggregate tree

        deepmerge(all_diffs_tree, diffs_tree)


else:
    all_diffs_tree = diffconfig(
        diff, None, from_filename, to_filename, output_filename, no_output,
        dump_rules, debug_parser, debug_convert)


# if the option to enable dumping of the differences tree is enabled, do
# that

if dump_diff_tree:
    print("=> differences tree:")
    print(yaml.dump(all_diffs_tree, default_flow_style=False))


exit(0)



# diff
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



# --- imports ---



from copy import deepcopy
import re
import sys

from deepops import (
    deepdiff, deepfilter, deepget, deepmerge, deepremoveitems, deepsetdefault)

import yaml



# --- functions ---



def pathstr(path, wildcard_indices=set()):
    """This function converts a path, which is a list of items of
    various types (typically strings and integers), into a string,
    useful for debugging messages.

    The items are separated by commas and elements which are None are
    converted to a '*' since they're wildcards.

    If the index of the item in the path is in the wildcard_indices
    list, that item is prefixed with "*=" to show it matched a wildcard.
    The default for this parameter is an empty list, which causes no
    index to match.

    If we just printed the list, it would by in the more verbose Python
    format, with surrounding square brackets, quotes around strings and
    None for a wildcard, which is a little less readable.
    """

    return ":".join([ ("*=" if i in wildcard_indices else "") + str(v)
                            for i, v in enumerate(path) ])



# --- classes ---



class DiffConvert:
    """This abstract class handles converting the difference between a
    'from' configuration item to the corresponding 'to' configuration
    item.

    The main difference process will use deepops.deepdiff() to work out
    what has been removed and what updated (added/changed) between the
    two configurations.

    Individual differences are checked using child classes, which
    specify the part of the configuration directionaries where they
    occur and the remove() and update() methods called.  For example, if
    the hostname is changed, a 'hostname' converter would specify the
    part of the configuration dictionary where the hostname is stored
    and the update() method would return the commands required to change
    it to the new value.

    The 'context', 'cmd' and 'ext' parameters, below, are all tuples
    giving parts of the path through the keys of the configuration
    dictionary to match for the specific converter.  These can be
    literal keys, or the value None, if that level of the path is a
    wildcard match for all keys - for example, '("interface", None)' to
    match all interfaces.  The part of the configuration dictionary
    that matches the 'context + cmd' parts of the path will be passed
    as the 'old/new' and 'rem/upd' parameters to the action methods.

    The child classes can override the following values:

    context -- the path to the context containing the commands used in
    this conversion: if this context is removed in its entirety, the
    remove() action for this converter will not be called as the
    removal of the context will remove it; this variable defaults to
    the empty tuple, for a command at the top level

    cmd -- the path for this command in the configuration, following
    the context; this part of the dictionary will be passed in the
    'old/new' and 'rem/upd' parameters to the action methods; this
    variable defaults to None and must be defined by a child class

    ext -- an extension to path which must be matched, but this part is
    not used to further specify the part of the dictionaries passed to
    the action functions for the 'old/new' and 'rem/upd' parameters:
    this is useful when a higher level in the configuration dictionary
    is required in the action function

    name -- if there is more than one action acting at a particular path
    point, this can be defined to differentiate them: it's useful when
    using the insert_before parameter, but isn't needed if the order
    does not matter (except for ease of debugging, to clarify which
    converter is being called)

    insert_before -- if this rule needs to happen prior to the actions
    of another rule, this 2-tuple gives the (path, name) before the
    earlier rule; None means the order does not matter

    In addition the remove() and update() methods will likely need to be
    implemented.
    """


    context = tuple()
    cmd = None    # leave as None to cause an error if not defined in child
    ext = tuple()
    name = None
    insert_before = None


    def __init__(self):
        """The constructor just precalculates some details to optimise
        the repeated processing of the converter matches.
        """

        # calculate and store a few things, for efficiency
        self._path_full = self.context + self.cmd + self.ext
        self._context_len = len(self.context)
        self._path_len = len(self.context + self.cmd)

        # store the set of indices of wildcard elements of the path (we
        # need these to get the argument list to pass to the converter
        # action methods)
        self.wildcard_indices = {
            i for i, v in enumerate(self._path_full) if v is None }


    def _path_matches(self, d, path):
        """This method is used to recursively step along the
        paths.  It is initially called by full_matches() from the top of
        the path - see that for more information.
        """

        # if the path is finished, return a single result with an
        # empty list, as there are no more path elements
        #
        # note that this is different from an empty list (which
        # would mean no matches
        if not path:
            return [ [] ]

        # get the path head and tail to make things easier to read
        path_head, path_tail = path[0], path[1:]

        # if this element is not a type we can iterate or search
        # through (we've perhaps reached a value or a None), or the
        # dictionary is empty, there are no matches, so just return
        # an empty list (which will cause all higher levels in the
        # path that matched to be discarded, giving no resulsts)
        if not isinstance(d, (dict, list, set)) or (not d):
            return []

        # if the path element is None, we're doing a wildcard match
        if path_head is None:
            # initialise an empty list for all returned matching
            # results
            results = []

            # go through all the keys at this level in the dictonary
            for d_key in d:
                # are there levels below this one?
                if path_tail:
                    # yes - recursively get the matches from under
                    # this key
                    for matches in self._path_matches(d[d_key], path_tail):
                        # add this match to the list of matches,
                        # prepending this key onto the start of the
                        # path
                        results.append([d_key] + matches)
                else:
                    # no - just add this result (this is a minor
                    # optimisation, as well as avoiding an error by
                    # trying to index into a non-dict type), above
                    results.append([d_key])

            return results

        # we have a literal key to match - if it's found,
        # recursively get the matches under this level
        if path_head in d:
            return [ ([path_head] + matches)
                            for matches
                            in self._path_matches(d[path_head], path_tail) ]

        # we have no matches, return the empty list
        return []


    def full_matches(self, d):
        """This method takes a dictionary and returns any matches for
        the full path in it, as a list of paths; each path is, itself, a
        list of keys.

        If any of the elements of the path are None, this is treated as
        a wildcard and will match all keys at that level.

        If there are no matching entries, the returned list will be
        empty.

        The returned list of matches is not guaranteed to be in any
        particular order.
        """

        return self._path_matches(d, self._path_full)


    def context_removed(self, d, match):
        """This method takes a remove dictionary, as returned by
        deepdiff(), and a specific match into it (which must be from
        the same DiffConvert object as the method is called on) and
        returns True iff the context for this converter is in it
        entirely and exactly.
        """

        # if this converter has no containing context, we definitely
        # can't be removing it, so will need to remove this item
        if not self.context:
            return False

        # get the part of the match fitting the context for this
        # converter
        match_context = match[0 : self._context_len]

        # see if the context is matched in its entirety in the remove
        # dictionary - if it's not empty, or we get an error doing it,
        # we're not removing the context
        #
        # KeyError indicates item not found; TypeError if level is not
        # a dictionary
        try:
            return deepget(d, *match_context, default_error=True) is None
        except (KeyError, TypeError):
            return False


    def remove(self, old, *args):
        """The remove() method is called when the specified path is in
        the remove differences (i.e. it's in the 'from' configuration
        but not in the 'to' configuration), unless the containing
        context is being removed in its entirety.

        The default behaviour of this method is to call truncate() with
        the old value for 'old' and 'remove' as, if there is no specific
        behaviour for removing an entire object, we should remove the
        individual elements.

        Keyword arguments:

        old -- the value of the dictionary item at the matching path in
        the 'from' configuration dictionary (sometimes, the full details
        of the old configuration may be required to remove it)

        *args -- [only] the arguments in the path which were matched
        with wildcards, supplied in the order they were given in the
        context, cmd and ext (e.g. for a match of ["interface", None,
        "standby", None], this will be second and fouth elements of
        the path); if there are no wildcards, no extra arguments will
        be supplied

        The return value is the commands to insert into the
        configuration to convert it.  This can either be a simple
        string, in the case of only one line being required, or an
        iterable containing one string per line.  If the return value is
        None, it is an indication nothing needed to be done.  An empty
        string or iterable indicates the update did something, which did
        not change the configuration (this may have semantic
        differences).
        """

        return self.truncate(old, old, *args)


    def truncate(self, old, rem, *args):
        """The truncate() method is identical to remove() except that it
        is called when an item is partially removed (i.e. something
        remains in the 'to' configuration - it is 'truncated').

        It is useful when amending lists or sets and matching at the
        containing object level.

        Keyword arguments:

        old -- the value of the dictionary item at the matching path in
        the 'from' configuration dictionary (sometimes, the full details
        of the old configuration may be required to remove it)

        rem -- the value of the dictionary item at the matching path in
        the remove differences dictionary (sometimes, it's necessary to
        know only what is removed from the new configuration)

        *args -- [only] the arguments in the path which were matched
        with wildcards, supplied in the order they were given in the
        context, cmd and ext (e.g. for a match of ["interface", None,
        "standby", None], this will be second and fouth elements of
        the path); if there are no wildcards, no extra arguments will
        be supplied

        The return value is the commands to insert into the
        configuration to convert it.  This can either be a simple
        string, in the case of only one line being required, or an
        iterable containing one string per line.  If the return value is
        None, it is an indication nothing needed to be done.  An empty
        string or iterable indicates the update did something, which did
        not change the configuration (this may have semantic
        differences).
        """

        pass


    def add(self, new, *args):
        """The add() method is called when the specified path is in the
        update differences but did not exist in the 'from' configuration
        (i.e. it's something new that is being added to the
        configuration).

        By default, this calls update() with the new configuration as
        the updated and new configuration arguments (and 'args') as, in
        many cases, the process for adding something is the same as
        updating it (e.g. updating an interface description).  It can,
        however, be overridden to do something different or, more
        commonly, implement add() but not update() for a particular
        change (as the updates will be picked up by more specific
        paths in other objects).

        Keyword arguments:

        new -- the value of the dictionary item at the matching path in
        the 'to' configuration dictionary

        *args -- [only] the arguments in the path which were matched
        with wildcards, supplied in the order they were given in the
        context, cmd and ext (e.g. for a match of ["interface", None,
        "standby", None], this will be second and fouth elements of
        the path); if there are no wildcards, no extra arguments will
        be supplied

        The return value is the commands to insert into the
        configuration to convert it.  This can either be a simple
        string, in the case of only one line being required, or an
        iterable containing one string per line.  If the return value is
        None, it is an indication nothing needed to be done.  An empty
        string or iterable indicates the update did something, which did
        not change the configuration (this may have semantic
        differences).
        """

        return self.update(None, new, new, *args)


    def update(self, old, upd, new, *args):
        """The update() method is called when the specified path is in
        the update differences and also in the 'from' configuration
        (i.e. something is being updated in the configuration).

        Keyword arguments:

        old -- the value of the dictionary item at the matching path in
        the 'from' configuration dictionary (sometimes, the old
        configuration must be removed first, before the new
        configuration can be updated, so the details are required)

        upd -- the value of the dictionary item at the matching path in
        the update differences dictionary

        new -- the value of the dictionary item at the matching path in
        the 'to' configuration dictionary (sometimes, the full details
        of the new configuration may be required to update it)

        *args -- [only] the arguments in the path which were matched
        with wildcards, supplied in the order they were given in the
        context, cmd and ext (e.g. for a match of ["interface", None,
        "standby", None], this will be second and fouth elements of
        the path); if there are no wildcards, no extra arguments will
        be supplied

        The return value is the commands to insert into the
        configuration to convert it.  This can either be a simple
        string, in the case of only one line being required, or an
        iterable containing one string per line.  If the return value is
        None, it is an indication nothing needed to be done.  An empty
        string or iterable indicates the update did something, which did
        not change the configuration (this may have semantic
        differences).
        """

        pass


    def get_path(self):
        """Get the non-extended path for this converter (i.e. the
        'context' and 'cmd' paths joined, excluding the 'ext').
        """

        return self.context + self.cmd


    def get_cfg(self, cfg, match):
        """This method returns the configuration to be passed to the
        converter's action methods [remove() and update()].

        By default, it indexes through the configuration using the path
        in the converter.  Converters may wish to override this, in some
        cases, for performance (perhaps if the entire configuration is
        to be returned).

        If the specified match path is not found, or there was a problem
        indexing along the path, None is returned, rather than an
        exception raised.
        """

        # try to get the specified matching path
        try:
            return deepget(cfg, *match[0 : self._path_len])

        except TypeError:
            # we got a TypeError so make the assumption that we've hit
            # a non-indexable element (such as a set) as the final
            # element of the path, so just return None
            return None


    def get_ext(self, cfg):
        """This method gets the extension part of the path, given a
        configuration dictionary starting at the path (i.e. what is
        passed as 'old/new' and 'rem/upd' in the action methods).

        An action method [remove() or update()] can use this to get the
        extension portion without needing to explicitly index through
        it.
        """

        return deepget(cfg, *self.ext)


    def get_wildcard_args(self, match):
        """This method returns a list of the wildcarded parts of the
        specified match as a list.  For example, if the path of this
        converter is 'interface', None) and the match is ['interface',
        'Vlan100'], the returned result will be ['Vlan100'].  This is
        used to supply the 'args' parameter to action methods.
        """

        return [ match[i] for i in self.wildcard_indices ]



class DiffConfig:
    """This abstract class is used to represent a configuration
    difference processor that can convert a configuration from one to
    another, using a method (which can be called once for each pair of
    configuration files).

    It encapsulates the rules for exluding items from the comparison.
    """


    def __init__(self, init_explain=False, init_dump_config=False,
                 init_dump_diff=False, init_debug_convert=0):

        """The constructor initialises the exclude list to empty and
        adds and converters using _add_converters().  It also stores
        some settings controlling the level of information describing
        the conversion process, based on the command line arguments:

        init_explain=False -- include comments in the output
        configuration changes that explain the differences being matched
        by the DiffConvert objects (if available).

        init_dump_config=False -- dump the 'from' and 'to'
        configurations (after exludes)

        init_dump_diff=False -- dump the differences (remove and update
        configurations)

        init_debug_convert=0 -- level of debugging information for the
        conversion process: >= 1: include steps, >= 2: include 'old/new'
        and 'rem/upd' parameters
        """

        # store the [initial] settings for the conversion process
        self._explain = init_explain
        self._dump_config = init_dump_config
        self._dump_diff = init_dump_diff
        self._debug_convert = init_debug_convert

        # initialise the dictionary of excludes, which will be passed
        # to deepremoveitems()
        self.init_rules()

        # initialise the list of converters and add them
        self._cvts = []
        self._add_converters()


    def _add_converters(self):
        """This method adds the converters for a particular object to
        the list used by the convert() method, usually by calling
        _add_converter() for each (see that method).

        The base class does nothing but child classes will implement it
        as they require.
        """

        pass


    def _add_converter(self, cvt):
        """Add an individual convert object, a child of the
        DiffConverter class, to the list of converters to be used by the
        convert() method.

        If the converter has an 'insert_before' entry that is not None,
        the converter will be inserted in the postition immediately
        before the existing entry with that name.  If that entry cannot
        be found, a KeyError will be raised.
        """

        if not cvt.insert_before:
            # no 'insert_before' - just add it to the end
            self._cvts.append(cvt)

        else:
            # 'insert_before' so we need to find that entry

            # get the indexes and entries in the converter list
            for pos, chk_cvt in enumerate(self._cvts):
                # if this entry is the one we're inserting before, put
                # it here and stop
                if chk_cvts.name == cvts.insert_before:
                    self._cvts.insert(pos, cvt)
                    break

            else:
                # we got the end of the list and didn't find it, so
                # raise an exception
                raise KeyError("cannot find existing converter with name: "
                               + cvt.insert_before)


    def _explain_comment(self, path):
        """This method returns a comment or other configuration item
        explaining the path supplied (which will typically be a match
        against a converter).  The path is supplied a list of levels and
        converted to a string using pathstr().

        Child classes can override this to provide a comment appropriate
        for their platform.

        If the function returns None, no comment is inserted.
        """

        return None


    def _diffs_begin(self):
        """This method returns a head (beginning) for a generated
        changes configuration file as an iterable of strings.

        In the abstract class, it returns None (which does not add
        anything) but, in child classes it may return a beginning line
        or similar.

        Note that if there are no changes, this will not be included in
        an otherwise empty file.
        """

        return []


    def _diffs_end(self):
        """This method returns a tail (ending) for a generated changes
        configuration file as an iterable of strings.

        In the abstract class, it returns None (which does not add
        anything) but, in child classes it may return an 'end' line or
        similar.

        Note that if there are no changes, this will not be included in
        an otherwise empty file.
        """

        return []


    def init_rules(self):
        """This method initialises the rules s.  In the base
        class, it the resets it to an empty dictionary, but child
        classes can extend this to add some default exclusions for
        standard system configuration entries which should not be
        changed.

        It is normally only called by the constructor but can be called
        later, to clear the excludes list before adding more (if
        required).
        """

        self.init_rules_tree()
        self.init_rules_active()


    def init_rules_tree(self):
        """This method initialises the rules tree (the dictionary of
        rules typically read from a file.

        In the base class, it just sets it to an empty dictionary but
        some platform-specific classes may wish to extend this to set up
        a default tree (along with init_rules_active()).
        """

        self._rules_tree = {}


    def init_rules_active(self):
        """This method initialises the active rules list (the list of
        rules specifying what should be used from the rules tree).

        In the base class, it just sets it to an empty list but some
        platform-specific classes may wish to extend this to set up a
        default list.
        """

        self._rules_active = []


    def add_rules_tree_file(self, filename):
        """Read a tree of rules items from a YAML file.  This is
        typically done once but then different portions of the rules
        dictionary selected with set_rules_active().

        The contents of the file are added to the current tree.  To
        clear the current tree first, use init_rules_tree().
        """

        try:
            file_rules_tree = yaml.safe_load(open(filename))

        except yaml.parser.ParserError as exception:
            raise ValueError("failed parsing rules file: %s: %s"
                                 % (filename, exception))

        self._rules_tree.update(file_rules_tree)


    def add_rules_active(self, rule_specs, devicename):
        """Add a list of rules to the current rule list, which specifies
        what parts of the rules tree should be used and how (include or
        exclude these items), to control the comparison done by the
        convert() method.

        The rules are specified as a list of strings in the format
        '[!]<path>' where '!' means 'exclude' (if omitted, it means
        'include') and 'path' is a colon-separated list of keys giving
        the path into the rules tree.  A particular element can be
        given as '%', in which case the 'devicename' parameter will be
        used but, if the devicename argument is None/empty (False),
        then a warning is printed and the rule skipped.

        For example, if '!device-excludes:%' is given, and the
        devicename is 'router1', the part of the rules tree indexed by
        ["device-excludes"]["router1"] will be excluded from the
        comparison.

        The rules given will be added to the current rules list; if the
        list is to be cleared first, use init_rules_list().
        """

        for rule_spec in rule_specs:
            # find if this is an include or exclude rule and get the
            # path part

            include = not rule_spec.startswith("!")

            if include:
                rule_spec_path = rule_spec
            else:
                rule_spec_path = rule_spec[1:]


            path_elements = rule_spec_path.split(":")

            if ("%" in path_elements) and (not devicename):
                print("warning: rule specification: %s contains '%%' but no "
                      "device name - ignoring" % rule_spec_path,
                      file=sys.stderr)

                continue


            path = [ devicename if i == "%" else i for i in path_elements ]

            self._rules_active.append( (include, path) )


    def get_rules_tree(self):
        """The method returns the current rules tree (as initialised by
        init_rules_tree() and extended with read_rules_tree()).  This
        should not be modified.
        """

        return self._rules_tree


    def get_rules(self):
        """This method just returns the active rules list and the
        portion of the rules tree it applies to.

        The return value is a list, one entry per rule, containing
        2-tuples: the first entry is the rule specification (a string,
        in the format described by add_rules_active()), and the second
        the portion of the rules tree that the path references (or None,
        if that part of the tree does not exist).

        This method is mainly used for debugging messages.
        """

        r = []
        for include, path in self._rules_active:
            r.append(
                (("" if include else "!") + ":".join(path),
                 deepget(self._rules_tree, *path) ) )

        return r


    def apply_rules(self, d):
        """This method applies the current rules (tree and active list)
        to the supplied configuration dictionary, either only including
        what's specified (using deepfilter()) or excluding (using
        deepremoveitems()) the specified portions.

        The configuration dictionary is modified in place.

        The method can be used on a configuration dictionary, or the
        remove/update dictionaries returned by deepdiff().
        """

        for include, path in self._rules_active:
            # get the porttion of the rules tree specified by the path
            # in this rule
            path_dict = deepget(self._rules_tree, *path)

            # skip to the next entry, if this part was not found
            if path_dict is None:
                continue

            # do either the include or exclude on the configuration
            # dictionary, in place)
            if include:
                d = deepfilter(d, path_dict)
            else:
                deepremoveitems(d, path_dict)


    def convert(self, from_cfg, to_cfg):
        """This method processes the conversion from the 'from'
        configuration to the 'to' configuration, removing excluded parts
        of each and calling the applicable converters' action methods.

        Note that, if excludes are used, the configurations will be
        modified in place by a deepremoveitems().  They will need to be
        copy.deepcopy()ed before passing them in, if this is
        undesirable.

        The returned value is a 2-tuple:

        - the first element is a big string of all the configuration
          changes that need to be made, sandwiched between
          _diffs_begin() and _diffs_end(), or None, if there were no
          differences

        - the second element is a dictionary giving the tree of
          differences (i.e. the elements where a difference was
          encountered - either a remove or an update)
        """


        self.apply_rules(from_cfg)

        if self._dump_config:
            print(">> 'from' configuration (after rules, if specified):",
                  yaml.dump(dict(from_cfg), default_flow_style=False),
                  sep="\n")


        # initialise the list of diffs (the returned configuration
        # conversions) and the tree of differences to empty

        diffs = []
        diffs_tree = {}


        # if no 'to' config was specified, stop here (we assume we're
        # just testing, parsing and excluding items from the 'from'
        # configuration and stopping
        #
        # we check for None explicitly for the difference between no
        # configuration and an empty configuration

        if to_cfg is None:
            return None, diffs_tree


        self.apply_rules(to_cfg)

        if self._dump_config:
            print(">> 'to' configuration (after rules, if specified):",
                yaml.dump(dict(to_cfg), default_flow_style=False), sep="\n")


        # use deepdiff() to work out the differences between the two
        # configuration dictionaries - what must be removed and what
        # needs to be added or updated
        #
        # then use deepfilter() to get the full details of each item
        # being removed, rather than just the top of a subtree being
        # removed

        remove_cfg, update_cfg = deepdiff(from_cfg, to_cfg)
        remove_cfg_full = deepfilter(from_cfg, remove_cfg)


        if self._dump_diff:
            print("=> differences - remove:",
                  yaml.dump(dict(remove_cfg), default_flow_style=False),

                  "=> differences - remove full:",
                  yaml.dump(dict(remove_cfg_full), default_flow_style=False),

                  "=> differences - update (add/change):",
                  yaml.dump(dict(update_cfg), default_flow_style=False),

                  sep="\n")


        # go through the list of converter objects

        for cvt in self._cvts:
            if self._debug_convert:
                print(">> checking difference:",
                      pathstr([ "*" if i is None else i
                                    for i in cvt.get_path() ]),
                      "name:", cvt.name)


            # get all the remove and update matches for this converter
            # and combine them into one list, discarding any duplicates
            #
            # we do this rather than processing each list one after the
            # other so we combine removes and updates on the same part
            # of the configuration together
            #
            # we combine them by manually building a list, rather than
            # using a set as a set requires hashable types

            remove_matches = cvt.full_matches(remove_cfg_full)
            update_matches = cvt.full_matches(update_cfg)

            all_matches = []
            for match in sorted(remove_matches + update_matches):
                if match not in all_matches:
                    all_matches.append(match)


            for match in all_matches:
                # handle REMOVE conversions, if matching

                if match in remove_matches:
                    if self._debug_convert:
                        print("=> remove match:",
                              pathstr(match, cvt.wildcard_indices))


                    # if we're removing the entire context for this
                    # match, we don't need to do it, as that will
                    # remove this, while it's at it

                    if cvt.context_removed(remove_cfg, match):
                        if self._debug_convert:
                            print("-> containing context being removed - skip")

                        continue


                    # check if anything remains at this level or below
                    # in the 'to' configuration - if it does, we're
                    # doing a partial removal

                    remove_is_trunc = False

                    try:
                        # we don't want the result here, just to find
                        # out if the path exists (if not, KeyError will
                        # be raised)
                        deepget(to_cfg, *match, default_error=True)

                    except KeyError:
                        # nothing remains so we're doing a full remove
                        pass

                    else:
                        # something remains so this is partial
                        if self._debug_convert:
                            print("-> subconfiguration not empty - truncate")
                        remove_is_trunc = True


                    # get the 'from' and 'remove' parts of the
                    # configuration and remove difference dictionaries,
                    # for the path specified in the converter (ignoring
                    # the extension 'ext')

                    cvt_old = cvt.get_cfg(from_cfg, match)
                    cvt_rem = cvt.get_cfg(remove_cfg_full, match)

                    if self._debug_convert >= 2:
                        print(
                            "-> from configuration:",
                            yaml.dump(cvt_old, default_flow_style=False),

                            "-> remove configuration:",
                            yaml.dump(cvt_rem, default_flow_style=False),

                            sep="\n")


                    # get elements in the path matching wildcards

                    args = cvt.get_wildcard_args(match)


                    # call the remove converter action method using the
                    # corresponding entry from the 'from' configuration

                    try:
                        if remove_is_trunc:
                            diff = cvt.truncate(cvt_old, cvt_rem, *args)
                        else:
                            diff = cvt.remove(cvt_old, *args)

                    except:
                        print("with: type(cvt)=" + repr(type(cvt)),
                              file=sys.stderr)

                        raise


                    # if some diffs were returned by the action, add
                    # them

                    if diff is not None:
                        # the return can be either a simple string or a
                        # list of strings - if it's a string, make it a
                        # list of one so we can do the rest the same way

                        if isinstance(diff, str):
                            diff = [diff]

                        if self._debug_convert:
                            print("\n".join(diff))


                        # add a comment, explaining the match, if
                        # enabled

                        if self._explain:
                            comment = self._explain_comment(match)
                            if comment:
                                diffs.append(comment)


                        # store this diff on the end of the list of
                        # diffs so far

                        diffs.extend(diff)
                        diffs.append("")


                        # add this match to the differences tree

                        deepsetdefault(diffs_tree, *match)


                    else:
                        if self._debug_convert:
                            print("-> no action")


                # handle UPDATE conversions, if matching

                if match in update_matches:
                    if self._debug_convert:
                        print("=> update match:",
                              pathstr(match, cvt.wildcard_indices))


                    args = cvt.get_wildcard_args(match)


                    # check if there is anything for this level in the
                    # 'from' configuration - if not, we're actually
                    # adding this, rather than updating it, so record
                    # that

                    update_is_add = False

                    try:
                        deepget(from_cfg, *match, default_error=True)

                    except KeyError:
                        if self._debug_convert:
                            print("-> no old configuration - add")
                        update_is_add = True


                    # get the 'from' and 'update' parts of the
                    # configuration and update difference dictionaries,
                    # for the path specified in the converter (ignoring
                    # the extension 'ext')

                    cvt_old = cvt.get_cfg(from_cfg, match)
                    cvt_upd = cvt.get_cfg(update_cfg, match)
                    cvt_new = cvt.get_cfg(to_cfg, match)

                    if self._debug_convert >= 2:
                        print(
                            "-> from configuration:",
                            yaml.dump(cvt_old, default_flow_style=False),

                            "-> update configuration:",
                            yaml.dump(cvt_upd, default_flow_style=False),

                            "-> to configuration:",
                            yaml.dump(cvt_new, default_flow_style=False),

                            sep="\n")


                    # call the update converter action method

                    try:
                        # if we're adding this, call the add() method,
                        # else update()
                        if update_is_add:
                            diff = cvt.add(cvt_new, *args)
                        else:
                            diff = cvt.update(cvt_old, cvt_upd, cvt_new, *args)

                    except:
                        print("with: type(cvt)=" + repr(type(cvt)),
                              file=sys.stderr)

                        raise


                    # (same as in remove, above)

                    if diff is not None:
                        if isinstance(diff, str):
                            diff = [diff]

                        if self._debug_convert:
                            print("\n".join(diff))


                        if self._explain:
                            comment = self._explain_comment(match)
                            if comment:
                                diffs.append(comment)


                        diffs.extend(diff)
                        diffs.append("")

                        deepsetdefault(diffs_tree, *match)

                    else:
                        if self._debug_convert:
                            print("-> no action")


            # print a blank line if debugging, just to space things out
            # nicely

            if self._debug_convert:
                print()

        # if nothing was generated, just return nothing

        if not diffs:
            return None, diffs_tree


        # return the diffs concatenated into a big, multiline string,
        # along with the begin and end blocks

        return ("\n".join(self._diffs_begin() + diffs + self._diffs_end()),
                diffs_tree)



# ios.utils
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Cisco IOS utility functions.

This module contains functions to help parsing Cisco IOS configuration
files.
"""



# --- imports ---



import difflib
import re

from netaddr import IPNetwork

from net_contextdiff.config import (
    IndentedContextualCommand, IndentedContextualConfig)



# --- interface functions ---



# interface names and their canonical representations - the canonical
# versions don't have to be how Cisco IOS always displays them, in any
# particular context, but just so they can consistently matched against

_INTERFACE_CANONICALS = {
    "ethernet": "Eth",
    "fastethernet": "Fa",
    "gigabitethernet": "Gi",
    "tengigabitethernet": "Te",
    "fortygigabitethernet": "Fo",
    "port-channel": "Po",
    "vl": "Vlan"
}



# regular expression for matching interface names and separating the
# time (e.g. 'Ethernet') from the number (e.g. '1' or '1/1.1')

_INTERFACE_NAME_RE = re.compile(r"([-a-z]+)([0-9/.]*)$")



def interface_canonicalize(name):
    """This function takes an interface name (including type and
    number) and returns a canonical version of it, handling
    abbreviations or case differences.

    At the moment, this just handles a defined set of names but it
    should really be more general and handle arbitrary short versions.
    """

    match = _INTERFACE_NAME_RE.match(name.lower())

    if match:
        type_, num = match.groups()
        type_ = _INTERFACE_CANONICALS.get(type_, type_).capitalize()
        name = type_ + num

    return name



# --- ACL functions ---



def _ipnet4_to_ios(ipnet4):
    """This function converts an IPv4 IPNetwork object into its
    canonical representation in Cisco IOS in a standard access-list.

    Conversions are '0.0.0.0/0' to 'any' and single host networks (i.e.
    simple addresses) into a plain address with no mask.
    """

    if ipnet4 == IPNetwork("0.0.0.0/0"):
        return "any"

    if ipnet4.prefixlen == 32:
        return str(ipnet4.ip)

    return str(ipnet4.ip) + " " + str(ipnet4.hostmask)



# IPV4 STANDARD ACCESS CONTROL LIST RULES



# regular expression for matching an IPv4 standard ACL rule, compiled
# once for efficiency

_IP_ACL_STD_RULE_RE = re.compile(
    r"^"
    r"(?P<action>permit|deny)"
    r" +"

    # we match "0.0.0.0 255.255.255.255" as "any" because the
    # netaddr module assumes the mask is a netmask (= /32) rather
    # than a hostmask, in this case
    r"((?P<any>any|0\.0\.0\.0 255\.255\.255\.255)|"
        r"(?:host )?(?P<host>[0-9.]+)|"
        r"(?P<net>[0-9.]+) (?P<mask>[0-9.]+))"

    r"$")


def _ip_acl_std_rule_parse(rule):
    """Parse an IPv4 standard ACL rule, returning a 2-element tuple
    consisting of the action ('permit' or 'deny') and an IPNetwork
    object, specifying the address/network to match.
    """


    match = _IP_ACL_STD_RULE_RE.match(rule)

    if not match:
        raise ValueError(
            "failed to parse IPv4 standard ACL rule: " + rule)


    # match some special cases and, if not those, match as a
    # network/hostmask

    if match.group("any"):
        ipnet4 = IPNetwork("0.0.0.0/0")

    elif match.group("host"):
        ipnet4 = IPNetwork(match.group("host"))

    else:
        ipnet4 = IPNetwork(match.group("net") + "/" + match.group("mask"))


    return match.group("action"), ipnet4



def ip_acl_std_canonicalize(l):
    """This function sorts an IPv4 standard access-list into a canoncial
    form so two lists can be compared.  The list is supplied and
    returned as a new list of IOS rules.  Note that it is not sorted in
    place), as parsed by _ip_acl_std_rule_parse().

    IPv4 standard ACLs are complicated in IOS due to its tendancy to
    reorganise the rules, after they're entered.  It always preserves
    the semantics of the list (never putting an overlapping 'permit' and
    'deny' in the wrong order), but it can move rules which don't
    interact around.  Presumably this is done to optimise processing.

    The solution adopted here is to build the ACLs up in blocks.  Each
    block is a set of rules where the address portions don't overlap;
    these are built, sorting each block into address order, before
    adding the sorted rules in the block into the resulting list.

    The result is lists which are not necessarily in the same order as
    they were constructed, nor how IOS stores them, but two lists should
    at least be in the same order so they can be directly compared.

    This function is applied to each list, after the configuration is
    read, and the returned list used to replace the order in which the
    rules were read.
    """


    # initialise the returned (canonical list)
    result = []

    # initialise the current block of non-overlapping rules
    block = []

    # go through the rules in the supplied list
    for rule in l:
        # parse the rule into action string and IPNetwork
        action, net = _ip_acl_std_rule_parse(rule)

        # find out if this rule overlaps with the network of a rule
        # already in the block
        overlap = [ None for _, chk_net
                         in block
                         if ((net.first <= chk_net.last)
                             and (net.last >= chk_net.first)) ]

        if overlap:
            # we had an overlap, add the current block to the returned
            # list and reinitialise it with just this new rule
            result.extend(block)
            block = [(action, net)]

        else:
            # add this rule to the current block and re-sort it on the
            # addresses of the rules in it
            block.append( (action, net) )
            block.sort(key=(lambda rule: rule[1]))

    # we've reached the end, so store the rules in the current block on
    # the end of the list
    result.extend(block)

    # convert the rules back into IOS text format and return them as a
    # list
    return [ (action + " " + _ipnet4_to_ios(net)) for action, net in result ]



# IPV4 EXTENDED ACCESS CONTROL LIST RULES



# regular expression for matching an IPv4 extended ACL rule, compiled
# once for efficiency

_IP_ACL_EXT_RULE_RE = re.compile(
    r"^"
    r"(?P<action>permit|deny)"
    r" +"
    r"(?P<protocol>ip|icmp|tcp|udp|igmp|pim|gre|esp)"
    r" "
    r"(?P<src_addr>any|host ([0-9.]+)|([0-9.]+) ([0-9.]+))"
    r"( ("
        # 'eq' and 'neq' can support a list of services - we need to match
        # them non-greedy
        r"((?P<src_port_listop>eq|neq) (?P<src_port_list>\S+( \S+)*?))|"

        r"((?P<src_port_1op>lt|gt) (?P<src_port_num>\S+))|"
        r"range (?P<src_port_low>\S+) (?P<src_port_high>\S+)"
    r"))?"
    r" "
    r"(?P<dst_addr>any|host ([0-9.]+)|([0-9.]+) ([0-9.]+))"
    r"( ("
        r"((?P<dst_port_listop>eq|neq) (?P<dst_port_list>\S+( \S+)*?))|"
        r"((?P<dst_port_1op>lt|gt) (?P<dst_port_num>\S+))|"
        r"range (?P<dst_port_low>\S+) (?P<dst_port_high>\S+)|"
        r"(?P<icmp_type>echo(-reply)?)"
    r"))?"
    r"(?P<established> established)?"
    r"(?P<qos> (dscp \S+))?"
    r"(?P<log> (log|log-input))?"
    r"$"
)



# _SERVICE_PORTS = dict
#
# Dictionary mapping service names (as displayed/usable in an access-
# list rule) into a port number.
#
# TODO: this list is not complete but parses all the rules we currently
# have in use.  It should be expanded at some point to all services.

_SERVICE_PORTS = {
    "bootps": 67,
    "bootpc": 68,
    "discard": 9,
    "domain": 53,
    "ftp": 21,
    "ftp-data": 20,
    "gopher": 70,
    "ident": 113,
    "isakmp": 500,
    "lpd": 515,
    "mail": 25,
    "non500-isakmp": 4500,
    "ntp": 123,
    "smtp": 25,
    "snmp": 161,
    "tftp": 69,
    "www": 80
}



def _port_canonicalize(service):
    """Converts a Cisco service named 'service' (as displayed/usable in
    an access-list rule) into a port number and return it as an
    integer.

    If the service name is not defined, it is assumed to already be
    numeric and is converted to an integer and returned.  If this
    conversion fails, an exception will be raised (whch probably
    indicates a service that is missing from the list).
    """

    return _SERVICE_PORTS.get(service) or int(service)



def ip_acl_ext_rule_canonicalize(rule):
    """Parse an IPv4 extended ACL rule, returning a 'normalised'
    form of the rule as a string.  The normalised form should allow
    two ACL rules which mean the same thing to be compared using a
    simple string comparison.

    This process mainly involves extracting the port entries and
    [potentially] translating them into port numbers, if they're named
    services (which can be used in rules, plus IOS will translate a
    numbered service to a name, if one matches).

    Note that no attempt is made to check the rule for validity.
    """


    match = _IP_ACL_EXT_RULE_RE.match(rule)

    if not match:
        raise ValueError(
            "failed to parse IPv4 extended ACL rule: " + rule)


    action, protocol, src_addr, dst_addr = match.group(
        "action", "protocol", "src_addr", "dst_addr")


    # match.group() will return an error if a named group does not exist
    # in the regexp; match.groupdict(), however, will return a default
    # value (None, if not specified) for named groups that do not exist
    #
    # as such, we need to check if the retrieved groups are blank or
    # not, for optional/alternative parts of a rule

    groups = match.groupdict()


    src_port = ""

    if groups["src_port_listop"]:
        # if 'eq' or 'neq' was found for the source port, it will be one
        # or more services, separated by spaces - we need to split the
        # list up and turn each one into a port number, then join the
        # list back together again

        src_port = (
            " %s %s" % (
                groups["src_port_listop"],
                " ".join([str(_port_canonicalize(s))
                              for s
                              in groups["src_port_list"].split(" ")])))

    elif groups["src_port_1op"]:
        src_port = " %s %d" % (
                       groups["src_port_1op"],
                       _port_canonicalize(groups["src_port_num"]))

    elif groups["src_port_low"]:
        src_port = " range %d %d" % (
                        _port_canonicalize(groups["src_port_low"]),
                        _port_canonicalize(groups["src_port_high"]))


    dst_port = ""

    if groups["dst_port_listop"]:
        dst_port = (
            " %s %s" % (
                groups["dst_port_listop"],
                " ".join([str(_port_canonicalize(s))
                              for s
                              in groups["dst_port_list"].split(" ")])))

    elif groups["dst_port_1op"]:
        dst_port = " %s %d" % (
                        groups["dst_port_1op"],
                        _port_canonicalize(groups["dst_port_num"]))

    elif groups["dst_port_low"]:
        dst_port = " range %d %d" % (
                        _port_canonicalize(groups["dst_port_low"]),
                        _port_canonicalize(groups["dst_port_high"]))

    # the destination port could also be an ICMP message type
    elif groups["icmp_type"]:
        dst_port = " " + groups["icmp_type"]


    established = groups["established"] or ""

    qos = groups["qos"] or ""

    log = groups["log"] or ""


    return (action + " " + protocol
            + " " + src_addr + src_port
            + " " + dst_addr + dst_port + established + qos + log)



# IPV6 EXTENDED ACCESS CONTROL LIST RULES



# regular expression for matching an IPv6 access-list rule, compiled
# once for efficiency
#
# TODO: we know this doesn't match some of the more complicated rules
# (such as the CP policing ones matching ICMPv6 types) but we're
# excluding those in the output, anyway, so we just ignore them - as
# such, we don't match the end of string ('$')

_IPV6_ACL_RULE_RE = re.compile(
    r"^"
    r"(?P<action>permit|deny)"
    r"( "
        r"(?P<protocol>ipv6|icmp|tcp|udp|\d+)"
    r")?"
    r" "
    r"(?P<src_addr>any|host [0-9A-Fa-f:]+|[0-9A-Fa-f:]+/\d+)"
    r"( ("
        r"((?P<src_port_1op>eq|lt|gt|neq) (?P<src_port_num>\S+))|"
        r"range (?P<src_port_low>\S+) (?P<src_port_high>\S+)"
    r"))?"
    r" "
    r"(?P<dst_addr>any|host [0-9A-Fa-f:]+|[0-9A-Fa-f:]+/\d+)"
    r"( ("
        r"((?P<dst_port_1op>eq|lt|gt|neq) (?P<dst_port_num>\S+))|"
        r"range (?P<dst_port_low>\S+) (?P<dst_port_high>\S+)|"
        r"(?P<icmp_type>echo(-reply)?)"
    r"))?"
    r"(?P<established> established)?"
    r"(?P<log> (log|log-input))?"
)


def ipv6_acl_rule_canonicalize(rule):
    """Parse an IPv6 ACL rule, returning a 'normalised' form of the rule
    as a string.  The normalised form should allow two ACL rules which
    mean the same thing to be compared using a simple string comparison.

    This process mainly involves extracting the port entries and
    [potentially] translating them into port numbers, if they're named
    services (which can be used in rules, plus IOS will translate a
    numbered service to a name, if one matches).

    Note that no attempt is made to check the rule for validity.
    """


    match = _IPV6_ACL_RULE_RE.match(rule)

    if not match:
        raise ValueError("failed to parse IPv6 ACL rule: " + rule)


    action, protocol, src_addr, dst_addr = match.group(
        "action", "protocol", "src_addr", "dst_addr")


    # if the protocol was not specified, we default to 'ipv6'

    if protocol is None:
        protocol = "ipv6"


    # lower case the source and destination addresses since IPv6
    # addresses can either be in upper or lower case (usually upper, in
    # IOS); we choose lower here, though, to avoid upper-casing the
    # keywords 'host' and 'any'

    src_addr = src_addr.lower()
    dst_addr = dst_addr.lower()


    # match.group() will return an error if a named group does not exist
    # in the regexp; match.groupdict(), however, will return a default
    # value (None, if not specified) for named groups that do not exist
    #
    # as such, we need to check if the retrieved groups are blank or
    # not, for optional/alternative parts of a rule

    groups = match.groupdict()


    src_port = ""

    if groups["src_port_num"]:
        src_port = " %s %d" % (
                       groups["src_port_1op"],
                       _port_canonicalize(groups["src_port_num"]))

    elif groups["src_port_low"]:
        src_port = " range %d %d" % (
                       _port_canonicalize(groups["src_port_low"]),
                       _port_canonicalize(groups["src_port_high"]))


    dst_port = ""

    if groups["dst_port_num"]:
        dst_port = " %s %d" % (
                       groups["dst_port_1op"],
                       _port_canonicalize(groups["dst_port_num"]))

    elif groups["dst_port_low"]:
        dst_port = " range %d %d" % (
                       _port_canonicalize(groups["dst_port_low"]),
                       _port_canonicalize(groups["dst_port_high"]))


    elif groups["icmp_type"]:
        dst_port = " " + groups["icmp_type"]


    established = groups["established"] or ""

    log = groups["log"] or ""


    return (action + " " + protocol
            + " " + src_addr + src_port
            + " " + dst_addr + dst_port + established)



# --- other functions ---



def expand_set(s):
    """This function exapands a string giving a set of numbers,
    separated by commas, which can include ranges from low to high,
    using hyphens.  The return value will be the set of numbers
    expressed.

    For example, given a string of "1,3-5", a set containing 1, 3, 4
    and 5 will be returned.
    """

    t = set()

    for i in s.split(","):
        i_range = i.split("-")

        if len(i_range) == 1:
            t.add(int(i))
        else:
            t.update(range(int(i_range[0]), int(i_range[1]) + 1))

    return t



# --- list differ ---



# create a differ object (we use it each time we do a comparison, so
# just create it once, at the start)
#
# this is only used by compare_lists(), below, so we don't need to
# export it

_differ = difflib.Differ()



def explain_diffs(old, new, indent="", prefix=""):
    """This function compares two lists (typically an access-list or
    prefix-list) and returns the commands to create the new list but
    interspersed with comments that explain the differences.

    The 'indent' string is added to the start of every line (including
    the comments) and is typically set to a space character, for 'ip
    access-list ...', where the rules are in a subcontext, but left the
    empty string for things like 'ip prefix-list ...', where they are
    not.

    The 'prefix' is added after the indent, before the text of each
    line.  It is not included on lines with comments.  It would be set
    to the empty string for 'ip access-list ...' (where no extra string
    is required) or something like 'ip prefix-list NAME ', where this
    is required on the same line.  Note that it must contain a trailing
    space, if one is required, between it and the line.

    If either of the two lists supplied are None, an empty list is
    assumed, either removing or adding all the entries.

    The return value is a list of lines these explaining comments
    inserted.  The lines without comments will create the new list.
    """

    # initialise the returned list
    r = []

    # keep track of if the last operation was to add an entry - we use
    # this to add markers at the beginning and end of an added block
    last_add = False

    # work through the differences between the two lists using
    # difflib.Differ.compare, which adds a character at the start,
    # explaining if a line is added, removed or the same, followed by
    # a space
    for line in _differ.compare(old or [], new or []):
        op, text = line[0], line[2:]

        if op == " ":
            # line is the same

            # if we were previously adding lines, indicate we're not
            # any more
            if last_add:
                r.append(indent + "!= ...")
                last_add = False

            # add this line
            r.append(indent + prefix + text)

        elif op == "+":
            # line is being added

            # if we're not already in a block of added lines, add a
            # comment, indicating that we're beginning doing that
            if not last_add:
                r.append(indent + "!+ ...")
                last_add = True

            # add this line
            r.append(indent + prefix + text)

        elif op == "-":
            # line is being removed

            # doesn't matter if we were already in a block, adding
            # lines or not, we're not doing that any more
            last_add = False

            # add a comment giving the line being removed
            r.append(indent + "!- " + prefix + text)

    # return the resulting list of lines
    return r



# --- other functions ---



# regular expression for matching physical interface names

_INTERFACE_PHYSICAL_RE = re.compile("^(Mgmt|Eth|Fa|Gi|Te|Fo)\d",
                                    re.IGNORECASE)


def is_int_physical(i):
    """This function returns if a particular interface is physical or
    not.  The names it matches are those that the are used as the
    canonical ones in the parser.
    """

    return _INTERFACE_PHYSICAL_RE.match(i)



# ios.config
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Cisco IOS configuration module.

This module parses Cisco IOS configuration files into a dictionary.
"""



# --- imports ---



import re

from deepops import deepsetdefault, deepget
from netaddr import IPNetwork

from net_contextdiff.config import IndentedContextualConfig

from .cmds import cmds
from .utils import ip_acl_std_canonicalize



# --- constants ---



# this dictionary specifies the settings in a portchannel interface
# which will propagate out to the member interfaces
#
# these settings are copied from the portchannel interface to the
# member interfaces, after a configuration is read

INTERFACE_PORTCHANNEL_MEMBER_CFG = [
    "storm-control",
    "switchport",
    "switchport-mode",
    "switchport-nonegotiate",
    "switchport-trunk-native",
    "switchport-trunk-allow",
]



# --- classes ----



class CiscoIOSConfig(IndentedContextualConfig):
    "This concrete class parses Cisco IOS configuration files."


    def _add_commands(self):
        """This method is called by the constructor to add commands for
        the IOS platform.

        The commands are stored in a global (to the module) level list
        of classes.
        """

        for cmd_class in cmds:
            self._add_command(cmd_class)


    def _post_parse_file(self):
        """Extend the inherited method to flush any pending IPv4
        standard ACL rules into the configuration.
        """

        super()._post_parse_file()


        # go through the pending IPv4 standard ACLs and store them (we
        # convert this to a list as _acl4_std_flush() will change the
        # dictionary during iteration and we only need the names)

        if "ip-access-list-standard" in self:
            ip_acl_std = self["ip-access-list-standard"]
            for name in ip_acl_std:
                ip_acl_std[name] = ip_acl_std_canonicalize(ip_acl_std[name])


        # go through the interfaces and copy settings from the
        # portchannel interface, if that interface is a member of one

        for int_name in self.get("interface", {}):
            # get the dictionary for this interface

            int_ = self["interface"][int_name]


            # if this interface is a portchannel member ...

            if "channel-group" in int_:
                id_ = int_["channel-group"][0]


                # get the corresponding portchannel interface this is a
                # member of

                po_int = self["interface"].get("Po%d" % id_, {})


                # copy the settings which propagate to the members into
                # this interface

                for setting in INTERFACE_PORTCHANNEL_MEMBER_CFG:
                    if setting in po_int:
                        int_[setting] = po_int[setting]



# ios.diff
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Cisco IOS configuration differences module.

This module compares two configurations for Cisco IOS devices and
produces a delta configuration.
"""



# --- imports ---



import re
import sys

import netaddr

from net_contextdiff.diff import DiffConvert, DiffConfig, pathstr
from .utils import is_int_physical, explain_diffs



# --- converter classes ---



# _cvts = []
#
# This is a list of converter classes to be added to the
# CiscoIOSConfigDiff object by the _add_converters() method.  Each
# class will be instantiated and added.

_cvts = []



# SYSTEM



class _Cvt_Hostname(DiffConvert):
    cmd = "hostname",

    def remove(self, old):
        return "no hostname"

    def update(self, old, upd, new):
        return "hostname " + upd

_cvts.append(_Cvt_Hostname)



# INTERFACE ...



class _Cvt_Int(DiffConvert):
    cmd = "interface", None

    def remove(self, old, int_name):
        # only remove the interface if it's not physical
        if is_int_physical(int_name):
            return

        return "no interface " + int_name

    def add(self, new, int_name):
        return "interface " + int_name

_cvts.append(_Cvt_Int)


class _CvtContext_Int(DiffConvert):
    "Abstract class for interface context converters to subclass."
    context = "interface", None


# we put the 'interface / shutdown' at the start to shut it down before
# we do any [re]configuration

class _Cvt_Int_Shutdown(_CvtContext_Int):
    cmd = "shutdown",
    name = "shutdown"

    def update(self, old, upd, new, int_name):
        return ("interface " + int_name,
                " " + ("" if upd else "no " ) + "shutdown")

_cvts.append(_Cvt_Int_Shutdown)


# we do VRF changes on an interface before we do any IP configuration,
# otherwise it will be removed

class _Cvt_Int_VRFFwd(_CvtContext_Int):
    cmd = "vrf-forwarding",

    def remove(self, old, int_name):
        return "interface " + int_name, " no vrf forwarding"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " vrf forwarding " + upd
        # TODO: need to find some way to trigger reapplication of IP
        # information (address, HSRP, etc.)

_cvts.append(_Cvt_Int_VRFFwd)


class _Cvt_Int_ARPTime(_CvtContext_Int):
    cmd = "arp-timeout",

    def remove(self, old, int_name):
        return "interface " + int_name, " no arp timeout"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " arp timeout %d" % upd

_cvts.append(_Cvt_Int_ARPTime)


class _Cvt_Int_CDPEna(_CvtContext_Int):
    cmd = "cdp-enable",

    def remove(self, old, int_name):
        # if the 'cdp enable' option is not present, that doesn't mean
        # it's disabled but just that it's not specified, so we assume
        # the default is for it to be enabled
        return "interface " + int_name, " cdp enable"

    def update(self, old, upd, new, int_name):
        return ("interface " + int_name,
                " " + ("" if upd else "no ") + "cdp enable")

_cvts.append(_Cvt_Int_CDPEna)


class _Cvt_Int_ChnGrp(_CvtContext_Int):
    cmd = "channel-group",

    def remove(self, old, int_name):
        return "interface " + int_name, " no channel-group"

    def update(self, old, upd, new, int_name):
        id_, mode = upd
        return ("interface " + int_name,
                " channel-group %d%s" % (id_, mode if mode else ""))

_cvts.append(_Cvt_Int_ChnGrp)


class _Cvt_Int_Desc(_CvtContext_Int):
    cmd = "description",

    def remove(self, old, int_name):
        return "interface " + int_name, " no description"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " description " + upd

_cvts.append(_Cvt_Int_Desc)


class _Cvt_Int_Encap(_CvtContext_Int):
    cmd = "encapsulation",

    def remove(self, old, int_name):
        return "interface " + int_name, " no encapsulation " + rem

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " encapsulation " + upd

_cvts.append(_Cvt_Int_Encap)


class _Cvt_Int_IPAccGrp(_CvtContext_Int):
    cmd = "ip-access-group", None

    def remove(self, old, int_name, dir_):
        return "interface " + int_name, " no ip access-group " + dir_

    def update(self, old, upd, new, int_name, dir_):
        return "interface " + int_name, " ip access-group %s %s" % (upd, dir_)

_cvts.append(_Cvt_Int_IPAccGrp)


class _Cvt_Int_IPAddr(_CvtContext_Int):
    cmd = "ip-address",

    def remove(self, old, int_name):
        return "interface " + int_name, " no ip address"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ip address " + upd

_cvts.append(_Cvt_Int_IPAddr)


class _Cvt_Int_IPAddrSec(_CvtContext_Int):
    cmd = "ip-address-secondary", None

    def remove(self, old, int_name, addr):
        return "interface " + int_name, " no ip address %s secondary" % addr

    def update(self, old, upd, new, int_name, addr):
        return "interface " + int_name, " ip address %s secondary" % addr

_cvts.append(_Cvt_Int_IPAddrSec)


class _Cvt_Int_IPFlowMon(_CvtContext_Int):
    cmd = "ip-flow-monitor", None

    def remove(self, old, int_name, dir_):
        return ("interface " + int_name,
                " no ip flow monitor %s %s" % (old, dir_))

    def update(self, old, upd, new, int_name, dir_):
        l = ["interface " + int_name]

        # we must remove the old flow monitor before setting a new one
        if old:
            l += [" no ip flow monitor %s %s" % (old, dir_)]

        l += [" ip flow monitor %s %s" % (upd, dir_)]
        return l

_cvts.append(_Cvt_Int_IPFlowMon)


class _Cvt_Int_IPHlprAddr(_CvtContext_Int):
    cmd = "ip-helper-address", None

    def remove(self, old, int_name, addr):
        return "interface " + int_name, " no ip helper-address " + addr

    def update(self, old, upd, new, int_name, addr):
        return "interface " + int_name, " ip helper-address " + addr

_cvts.append(_Cvt_Int_IPHlprAddr)


class _Cvt_Int_IPIGMPVer(_CvtContext_Int):
    cmd = "ip-igmp-version",

    def remove(self, old, int_name):
        return "interface " + int_name, " no ip igmp version"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ip igmp version " + upd

_cvts.append(_Cvt_Int_IPIGMPVer)


class _Cvt_Int_IPMcastBdry(_CvtContext_Int):
    cmd = "ip-multicast-boundary",

    def remove(self, old, int_name):
        return "interface " + int_name, " no ip multicast boundary"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ip multicast boundary " + upd

_cvts.append(_Cvt_Int_IPMcastBdry)


class _Cvt_Int_IPPIMMode(_CvtContext_Int):
    cmd = "ip-pim", "mode"

    def remove(self, old, int_name):
        return "interface " + int_name, " no ip pim " + old

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ip pim " + upd

_cvts.append(_Cvt_Int_IPPIMMode)


class _Cvt_Int_IPPIMBSRBdr(_CvtContext_Int):
    cmd = "ip-pim", "bsr-border"

    def remove(self, old, int_name):
        return "interface " + int_name, " no ip pim bsr-border"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ip pim bsr-border"

_cvts.append(_Cvt_Int_IPPIMBSRBdr)


class _Cvt_Int_IPProxyARP(_CvtContext_Int):
    cmd = "ip-proxy-arp",

    def update(self, old, upd, new, int_name):
        return ("interface " + int_name,
                " " + ("" if upd else "no ") + "ip proxy-arp")

_cvts.append(_Cvt_Int_IPProxyARP)


class _Cvt_Int_IPVerifyUni(_CvtContext_Int):
    cmd = "ip-verify-unicast",

    def remove(self, old, int_name):
        return "interface " + int_name, " no ip verify unicast"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ip verify unicast " + upd

_cvts.append(_Cvt_Int_IPVerifyUni)


class _Cvt_Int_IPv6Addr(_CvtContext_Int):
    cmd = "ipv6-address", None

    def remove(self, old, int_name, addr):
        return "interface " + int_name, " no ipv6 address " + addr

    def update(self, old, upd, new, int_name, addr):
        return "interface " + int_name, " ipv6 address " + addr

_cvts.append(_Cvt_Int_IPv6Addr)


class _Cvt_Int_IPv6MultBdry(_CvtContext_Int):
    cmd = "ipv6-multicast-boundary-scope",

    def remove(self, old, int_name):
        return "interface " + int_name, " no ipv6 multicast boundary scope"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ipv6 multicast boundary scope " + upd

_cvts.append(_Cvt_Int_IPv6MultBdry)


class _Cvt_Int_IPv6PIMBSRBdr(_CvtContext_Int):
    cmd = "ipv6-pim", "bsr-border"

    def remove(self, old, int_name):
        return "interface " + int_name, " no ipv6 pim bsr border"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ipv6 pim bsr border"

_cvts.append(_Cvt_Int_IPv6PIMBSRBdr)


class _Cvt_Int_IPv6TrafFilt(_CvtContext_Int):
    cmd = "ipv6-traffic-filter", None

    def remove(self, old, int_name, dir_):
        return "interface " + int_name, " no ipv6 traffic-filter " + dir_

    def update(self, old, upd, new, int_name, dir_):
        return ("interface " + int_name,
               " ipv6 traffic-filter %s %s" % (upd, dir_))

_cvts.append(_Cvt_Int_IPv6TrafFilt)


class _Cvt_Int_IPv6VerifyUni(_CvtContext_Int):
    cmd = "ipv6-verify-unicast",

    def remove(self, old, rem, int_name):
        return "interface " + int_name, " no ipv6 verify unicast"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " ipv6 verify unicast " + upd

_cvts.append(_Cvt_Int_IPv6VerifyUni)


class _Cvt_Int_ServPol(_CvtContext_Int):
    cmd = "service-policy",

    # this uses truncate() to ensure that all removes happen before any
    # update()s - this because an old policy must be removed before a
    # new one of the same type/direction is added
    #
    # calls to remove() will be matched in alphabetical order which
    # would match the add first, if it's earlier in the list

    def truncate(self, old, rem, int_name):
        l = ["interface " + int_name]
        for policy in sorted(rem):
            l.append(" no service-policy " + policy)
        return l

    def update(self, old, upd, new, int_name):
        l = ["interface " + int_name]
        for policy in sorted(upd):
            l.append(" service-policy " + policy)
        return l

_cvts.append(_Cvt_Int_ServPol)


class _Cvt_Int_StandbyIP(_CvtContext_Int):
    cmd = "standby", "group", None, "ip"

    def remove(self, old, int_name, grp):
        return "interface " + int_name, " no standby %d ip" % grp

    def update(self, old, upd, new, int_name, grp):
        return "interface " + int_name, " standby %d ip %s" % (grp, upd)

_cvts.append(_Cvt_Int_StandbyIP)


class _Cvt_Int_StandbyIPSec(_CvtContext_Int):
    cmd = "standby", "group", None, "ip-secondary", None

    def remove(self, old, int_name, grp, addr):
        return ("interface " + int_name,
                " no standby %d ip %s secondary" % (grp, addr))

    def update(self, old, upd, new, int_name, grp, addr):
        return ("interface " + int_name,
                " standby %d ip %s secondary" % (grp, addr))

_cvts.append(_Cvt_Int_StandbyIPSec)


class _Cvt_Int_StandbyIPv6(_CvtContext_Int):
    cmd = "standby", "group", None, "ipv6", None

    def remove(self, old, int_name, grp, addr):
        return ("interface " + int_name,
                " no standby %d ipv6 %s" % (grp, addr))

    def update(self, old, upd, new, int_name, grp, addr):
        return ("interface " + int_name,
                " standby %d ipv6 %s" % (grp, addr))

_cvts.append(_Cvt_Int_StandbyIPv6)


class _Cvt_Int_StandbyPreempt(_CvtContext_Int):
    cmd = "standby", "group", None, "preempt"

    def remove(self, old, int_name, grp):
        return "interface " + int_name, " no standby %d preempt" % grp

    def update(self, old, upd, new, int_name, grp):
        return "interface " + int_name, " standby %d preempt" % grp

_cvts.append(_Cvt_Int_StandbyPreempt)


class _Cvt_Int_StandbyPri(_CvtContext_Int):
    cmd = "standby", "group", None, "priority"

    def remove(self, old, int_name, grp):
        return "interface " + int_name, " no standby %d priority" % grp

    def update(self, old, upd, new, int_name, grp):
        return "interface " + int_name, " standby %d priority %d" % (grp, upd)

_cvts.append(_Cvt_Int_StandbyPri)


class _Cvt_Int_StandbyTimers(_CvtContext_Int):
    cmd = "standby", "group", None, "timers"

    def remove(self, old, int_name, grp):
        return "interface " + int_name, " no standby %d timers" % grp

    def update(self, old, upd, new, int_name, grp):
        return "interface " + int_name, " standby %d timers %s" % (grp, upd)

_cvts.append(_Cvt_Int_StandbyTimers)


class _Cvt_Int_StandbyTrk(_CvtContext_Int):
    cmd = "standby", "group", None, "track", None

    def remove(self, old, int_name, grp, obj):
        return "interface " + int_name, " no standby %d track %s" % (grp, obj)

    def update(self, old, upd, new, int_name, grp, obj):
        return ("interface " + int_name,
                " standby %d track %s%s"
                    % (grp, obj, (" " + upd) if upd else ""))

_cvts.append(_Cvt_Int_StandbyTrk)


class _Cvt_Int_StandbyVer(_CvtContext_Int):
    cmd = "standby", "version"

    def remove(self, old, int_name):
        return "interface " + int_name, " no standby version"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " standby version %d" % upd

_cvts.append(_Cvt_Int_StandbyVer)


class _Cvt_Int_StormCtrl(_CvtContext_Int):
    cmd = "storm-control", None

    def remove(self, old, int_name, traffic):
        return ("interface " + int_name,
                " no storm-control %s level" % traffic)

    def update(self, old, upd, new, int_name, traffic):
        return ("interface " + int_name,
                " storm-control %s level %.2f" % (traffic, upd))

_cvts.append(_Cvt_Int_StormCtrl)


class _Cvt_Int_SwPort(_CvtContext_Int):
    cmd = "switchport",

    def remove(self, old, int_name):
        # if the 'switchport' option is not present, that doesn't mean
        # it's disabled but just that it's not specified, so we assume
        # the default is for it to be disabled
        #
        # TODO: this is the case for routers (which we're concerned
        # about here) but not switches: we'd probably need a separate
        # platform for this
        return "interface " + int_name, " no switchport"

    def update(self, old, upd, new, int_name):
        return ("interface " + int_name,
                " " + ("" if upd else "no ") + "switchport")

_cvts.append(_Cvt_Int_SwPort)


class _Cvt_Int_SwPortMode(_CvtContext_Int):
    cmd = "switchport-mode",

    def remove(self, old, int_name):
        return "interface " + int_name, " no switchport mode"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " switchport mode " + upd

_cvts.append(_Cvt_Int_SwPortMode)


class _Cvt_Int_SwPortNoNeg(_CvtContext_Int):
    cmd = "switchport-nonegotiate",

    def remove(self, old, int_name):
        return "interface " + int_name, " no switchport nonegotiate"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " switchport nonegotiate"

_cvts.append(_Cvt_Int_SwPortNoNeg)


class _Cvt_Int_SwPortTrkNtv(_CvtContext_Int):
    # we just match the interface as we need to look inside it to see if
    # the interface is part of a channel group
    cmd = tuple()
    ext = "switchport-trunk-native",

    def remove(self, old, int_name):
        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in old:
            return None

        return "interface " + int_name, " no switchport trunk native vlan"

    def update(self, old, upd, new, int_name):
        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in new:
            return None

        return ("interface " + int_name,
                " switchport trunk native vlan %d" % self.get_ext(upd))

_cvts.append(_Cvt_Int_SwPortTrkNtv)


class _Cvt_Int_SwPortTrkAlw(_CvtContext_Int):
    # we just match the interface as we need to look inside it to see if
    # the interface is part of a channel group
    cmd = tuple()
    ext = "switchport-trunk-allow",

    def truncate(self, old, rem, int_name):
        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in old:
            return None

        l = ["interface " + int_name]
        for tag in sorted(self.get_ext(rem)):
            l.append(" switchport trunk allowed vlan remove %d" % tag)
        return l

    def update(self, old, upd, new, int_name):
        # if this interface is in a port-channel, we do all changes
        # there, so skip this
        if "channel-group" in new:
            return None

        l = ["interface " + int_name]
        for tag in sorted(self.get_ext(upd)):
            l.append(" switchport trunk allowed vlan add %d" % tag)
        return l

_cvts.append(_Cvt_Int_SwPortTrkAlw)


class _Cvt_Int_XConn(_CvtContext_Int):
    cmd = "xconnect",

    def remove(self, old, int_name):
        return "interface " + int_name, " no xconnect"

    def update(self, old, upd, new, int_name):
        return "interface " + int_name, " xconnect " + upd

_cvts.append(_Cvt_Int_XConn)


# we put the 'interface / no shutdown' at the end to only enable the
# interface once it's been correctly [re]configured

class _Cvt_Int_NoShutdown(_CvtContext_Int):
    cmd = "shutdown",
    name = "no-shutdown"

    def remove(self, old, int_name):
        return "interface " + int_name, " no shutdown"

_cvts.append(_Cvt_Int_NoShutdown)



# IP[V6] ACCESS-LIST ...



class _Cvt_IPACL_Std(DiffConvert):
    cmd = "ip-access-list-standard", None

    def remove(self, old, acl_name):
        return "no ip access-list standard " + acl_name

    def update(self, old, upd, new, acl_name):
        r = ["ip access-list standard " + acl_name]
        r.extend(explain_diffs(old, new, indent=" "))
        return r

_cvts.append(_Cvt_IPACL_Std)


class _Cvt_IPACL_Ext(DiffConvert):
    cmd = "ip-access-list-extended", None

    def remove(self, old, acl_name):
        return "no ip access-list extended " + acl_name

    def update(self, old, upd, new, acl_name):
        r = ["ip access-list extended " + acl_name]
        r.extend(explain_diffs(old, new, indent=" "))
        return r

_cvts.append(_Cvt_IPACL_Ext)


class _Cvt_IPv6ACL_Ext(DiffConvert):
    cmd = "ipv6-access-list", None

    def remove(self, old, acl_name):
        return "no ipv6 access-list " + acl_name

    def update(self, old, upd, new, acl_name):
        r = ["ipv6 access-list " + acl_name]
        r.extend(explain_diffs(old, new, indent=" "))
        return r

_cvts.append(_Cvt_IPv6ACL_Ext)



# IP[V6] PREFIX-LIST ...



class _Cvt_IPPfxList(DiffConvert):
    cmd = "ip-prefix-list", None

    def remove(self, old, pfx_name):
        return "no ip prefix-list " + pfx_name

    def update(self, old, upd, new, pfx_name):
        return explain_diffs(old, new, prefix="ip prefix-list %s " % pfx_name)

_cvts.append(_Cvt_IPPfxList)


class _Cvt_IPv6PfxList(DiffConvert):
    cmd = "ipv6-prefix-list", None

    def remove(self, old, pfx_name):
        return "no ipv6 prefix-list " + pfx_name

    def update(self, old, upd, new, pfx_name):
        return explain_diffs(
                   old, new, prefix="ipv6 prefix-list %s " % pfx_name)

_cvts.append(_Cvt_IPv6PfxList)



# IP[V6] ROUTE ...



class _Cvt_IPRoute(DiffConvert):
    cmd = "ip-route", None

    def remove(self, old, route):
        return "no ip route " + route

    def update(self, old, upd, new, route):
        return "ip route " + route

_cvts.append(_Cvt_IPRoute)


class _Cvt_IPv6Route(DiffConvert):
    cmd = "ipv6-route", None

    def remove(self, old, route):
        return "no ipv6 route " + route

    def update(self, old, upd, new, route):
        return "ipv6 route " + route

_cvts.append(_Cvt_IPv6Route)



# [NO] SPANNING-TREE ...



class _Cvt_NoSTP(DiffConvert):
    cmd = "no-spanning-tree-vlan", None

    def remove(self, old, tag):
        # removing 'no spanning-tree' enables spanning-tree
        return "spanning-tree vlan %d" % tag

    def update(self, old, upd, new, tag):
        # adding 'no spanning-tree' disables spanning-tree
        return "no spanning-tree vlan %d" % tag

_cvts.append(_Cvt_NoSTP)


class _Cvt_STPPri(DiffConvert):
    cmd = "spanning-tree-vlan-priority", None

    def remove(self, old, tag):
        return "no spanning-tree vlan %d priority" % tag

    def update(self, old, upd, new, tag):
        return "spanning-tree vlan %d priority %d" % (tag, upd)

_cvts.append(_Cvt_STPPri)



# TRACK ...



class _Cvt_Track(DiffConvert):
    cmd = "track", None
    ext = "criterion",

    def remove(self, old, obj):
        return "no track %d" % obj

    def update(self, old, upd, new, obj):
        return "track %d %s" % (obj, upd["criterion"])

_cvts.append(_Cvt_Track)


class _CvtContext_Track(DiffConvert):
    context = "track", None


class _Cvt_Track_Delay(_CvtContext_Track):
    cmd = "delay",

    def remove(self, old, obj):
        return "track %d" % obj, " no delay"

    def update(self, old, upd, new, obj):
        return "track %d" % obj, " delay " + upd

_cvts.append(_Cvt_Track_Delay)


class _Cvt_Track_IPVRF(_CvtContext_Track):
    cmd = "ip-vrf",

    def remove(self, old, obj):
        return "track %d" % obj, " no ip vrf"

    def update(self, old, upd, new, obj):
        return "track %d" % obj, " ip vrf " + upd

_cvts.append(_Cvt_Track_IPVRF)


class _Cvt_Track_IPv6VRF(_CvtContext_Track):
    cmd = "ipv6-vrf",

    def remove(self, old, obj):
        return "track %d" % obj, " no ipv6 vrf"

    def update(self, old, upd, new, obj):
        return "track %d" % obj, " ipv6 vrf " + upd

_cvts.append(_Cvt_Track_IPv6VRF)


class _Cvt_Track_Obj(_CvtContext_Track):
    context = "track", None
    cmd = "object", None

    def remove(self, old, obj, sub_obj):
        return "track %d" % obj, " no object " + sub_obj

    def update(self, old, upd, new, obj, sub_obj):
        return "track %d" % obj, " object " + sub_obj

_cvts.append(_Cvt_Track_Obj)



# VLAN ...



class _Cvt_VLAN(DiffConvert):
    cmd = "vlan", None

    def remove(self, old, tag):
        return "no vlan %d" % tag

    def add(self, new, tag):
        return "vlan %d" % tag

_cvts.append(_Cvt_VLAN)


class _Cvt_VLAN_Name(DiffConvert):
    context = "vlan", None
    cmd = "name",

    def remove(self, old, tag):
        return "vlan %d" % tag, " no name"

    def update(self, old, upd, new, tag):
        return "vlan %d" % tag, " name " + upd

_cvts.append(_Cvt_VLAN_Name)



# --- context parser ---



class CiscoIOSDiffConfig(DiffConfig):
    """This class is used to compare two IOS configuration files and
    generate a configuration file to transform one into the other.
    """


    def _add_converters(self):
        "This method adds the converters for Cisco IOS."

        for cvt_class in _cvts:
            try:
                self._add_converter(cvt_class())
            except:
                print("with: cvt_class=" + repr(cvt_class),
                      file=sys.stderr)

                raise


    def _explain_comment(self, path):
        """This method overrides the empty inherited one to return a
        Cisco IOS comment giving the matched path.
        """

        return "! => " + pathstr(path)


    def _diffs_end(self):
        """This method overrides the empty inherited one to return a
        single line saying 'end'.
        """

        return ["end"]


    def init_rules_tree(self):
        """This method extends the inherited one to add some rules to
        the tree for the default CoPP (Control Plane Policing) IPv4
        extended and IPv6 ACLs.
        """

        super().init_rules_tree()

        self._rules_tree.update( {
            "ios-default": {
                "ip-access-list-extended": {
                    "acl-copp-match-igmp": {},
                    "acl-copp-match-pim-data": {},
                },
                "ipv6-access-list": {
                    "acl-copp-match-mld": {},
                    "acl-copp-match-ndv6": {},
                    "acl-copp-match-ndv6hl": {},
                    "acl-copp-match-pimv6-data": {},
                },
            },
        } )


    def init_rules_active(self):
        """This method extends the inherited one to add in some active
        rules to exclude the portions of the rules tree set up in
        init_rules_tree().
        """

        super().init_rules_active()

        self._rules_active.append( (False, ("ios-default", ) ) )



# config
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Contextual configuration module.

This module contains an abstract class to parse and represent
configuration files that use indented blocks to indicate the context of
commands.
"""



# --- imports ---



import re
import sys



# --- classes ---



class IndentedContextualCommand:
    """This is an abstract class that represents a command from a
    configuration file using indented contexts.

    It includes the context in which it occurs (the indented block), a
    regular expression to match it (which will include groups to extract
    parameters from), if it enters a subcontext and an parse() method
    to run, if there is a match (which might store configuration
    information).

    Actual commands will inherit from this and override or implement
    some of the constants:

    context -- the context (indented section name) in which the command
    applies; the default is '.', which represents the top level (no
    indent, i.e. no subcontext)

    match -- an uncompiled regular expression string which fully matches
    (with re.fullmatch()) the command, case insensitively

    enter_context -- if not Null, it will be a string specifying the
    subcontext to enter, if the command matches

    In addtion, the parse() method is likely to need to be implemented.
    """


    context = "."
    match = None
    enter_context = None


    def __init__(self):
        """The constructor simply compiles the string-based regular
        expression and stores it once, for performance.
        """

        # if there's a mistake in a regular expression, the trace just
        # displays the commands and not which particular command was at
        # fault, so we catch that exception and display the command
        self._match_re = re.compile(self.match + r"$", re.IGNORECASE)


    def parse(self, cfg):
        """This method is called when the command is matched.  It will
        typically be used to update the configuration dictionary.

        The default is not to do anything - child classes will likely
        need to implement this.

        The configuration dictionary supplied to this function is not
        necessarily the entire configuration for the device but what is
        returned by the parse() method of the parent context.  This
        allows a parent context to pass on just the configuration
        dictionary for a portion of the device, such as an individual
        interface.  This makes it easier to update the correct portion
        of the configuration, without having to index through the entire
        structure.

        Keyword arguments:

        cfg -- the dictionary for this part of the configuration, or the
        the full configuration, as returned by the parent context's
        parse() method

        *groups -- the re.Match.groupdict() returned after matching the
        command regular expression will be expanded into keyword
        arguments as '**groupdict()', so there'll be one arguemnt for
        each named groups, with the corresponding name (e.g. if the
        regular expression contains '(<?P<addr>...)' an argument will
        be provided called 'addr' with the matching pattern)

        The return value is the dictionary to use as the 'cfg' argument
        of any subcontext parse() calls, or None if the same
        configuration dictionary as the parent context is to be used.
        """

        pass


    def ismatch(self, line):
        """This method returns if this command matches the supplied
        command line.  The regular expression must be matched
        in its entirity (it uses re.fullmatch()) and does not include
        any leading indenting due to the context.
        """

        return self._match_re.fullmatch(line)



class IndentedContextualConfig(dict):
    """The abstract base class for an indent-based contextual
    configuration.

    Indent-based configurations use levels of indenting to indicate
    context - e.g.:

      interface X
       description Y
       shutdown

    This shows that the 'description' and 'shutdown' commands apply to
    the interface named in the previous line.

    The parser in this class makes use of this property to match only
    the commands which are applicable in this context the configuration
    is in.  When the indenting level decreases, the contexts are exited
    and commands for the appropriate parent context will be matched.

    Child classes implement the platform-specific commands (e.g. one to
    handle Cisco IOS configuration files).
    """


    def __init__(self, d=None, filename=None, debug=0):
        """Initialise the object by calling _add_commands() to add the
        commands for this configuration platform.

        If an initial configuration dictionary is supplied, use that
        (this allows the constructor to behave like the inherited one
        from the 'dict' class).

        If a filename (and optional debug argument) are supplied, the
        parse_file() method is called with these.
        """


        # if an initial object to store as the configuration dictionary
        # was supplied, call the inherited constructor with that, else
        # just call it with no initial value

        if not d:
            super().__init__()
        else:
            super().__init__(d)


        # _parser = {}
        #
        # command parsing is handled through a dictionary indexed by
        # the name of the context and then having a list of Indented-
        # ContextualCommand objects for that context, to be processed in
        # turn
        #
        # the command dictionary is initialised with a single context
        # of '.' (the top level context) containing no commands - new
        # contexts will be created, as required, but we need to start
        # with this as the matching function assumes it exists

        self._parser = {
            ".": []
        }


        # add the commands for this particular instance

        self._add_commands()


        # if a filename was supplied, add its contents to the
        # configuration

        if filename:
            self.parse_file(filename, debug)


    def _add_commands(self):
        """This method should be implemented by child classes to add
        commands to the parser.

        In the abstract class, it does nothing.
        """

        pass


    def _add_command(self, cmd_class):
        """This method adds a single IndentedContextualCommand class to
        the parser, given the class: an object of that class will be
        instantiated and added to the appropriate context in the
        command dictionary.
        """

        # if the context doesn't exist yet, create it with an empty list
        # of commands

        self._parser.setdefault(cmd_class.context, [])


        # instantiate the command object and append it to this context's
        # command list

        try:
            self._parser[cmd_class.context].append(cmd_class())

        except:
            print("with: cmd_class=" + repr(cmd_class), file=sys.stderr)
            raise


    def parse_file(self, filename, debug=0):
        """Parse a device configuration file and stores it in the
        object.  If there is any configuration already in the object,
        it will be added to it.

        Keyword arguments:

        filename -- name of the file to read

        debug -- if set to True, explain the parsing process, showing
        the contexts and matching commands
        """


        file = open(filename, "r")


        # the context stack stores the contexts for each level of
        # indentation in the file so far: as contexts are entered, they
        # are pushed onto it and, as they are left, they are popped off
        #
        # the top of the stack represents the current context
        #
        # each level is a dictionary with the following keys:
        #
        # * indent - the number of leading spaces for the command which
        #   started this context (so commands within it will have more
        #   than this number of spaces)
        #
        # * name - the name of the context (as matched in Indented-
        #   ContextualCommand)
        #
        # * config - the configuration for the current context
        #
        # the list is initialised with the top level context (no indent)

        contexts = [ {
            # the top level has a negative indent so all commands are
            # within in
            "indent": -1,

            # the top level has the special name '.'
            "name": ".",

            # the configuration starts by being the entire dictionary
            "config": self,
        } ]


        for line in file:
            # skip blank lines (or lines consisting solely of spaces)

            if not line.lstrip():
                continue


            # remove the trailing newline (but not any trailing spaces,
            # in case they're important)

            line = line.rstrip("\n")


            if debug:
                print("\n" + line)


            # strip the leading spaces from the line and calculate the
            # indent level based on this

            line_strip = line.lstrip(" ")
            line_indent = len(line) - len(line_strip)


            # keep popping contexts until the level of indent of the
            # line is greater than the top context on the stack (in
            # which case, we've found the context this command is in)

            while contexts[-1]["indent"] >= line_indent:
                contexts.pop()


            # get the current context (we need this a lot, so it makes
            # the code clearer to do it here)

            context = contexts[-1]


            # start building a dictionary for the new context with
            # default values - we may replace some of these, if we find
            # a matching command

            new_context = {
                # the indent level of this context is that of this line
                "indent": line_indent,

                # the new context doesn't yet have name (and may not
                # have one, if we can't find one)
                "name": None,

                # keep the same configuration element
                "config": context["config"]
            }


            if debug:
                print(">> context:", context["name"] or "(none)")


            # skip to the next line, if the current context if it has no
            # name specified (so there can't be any commands)
            #
            # note that we only switch to a context if there are any
            # commands registered in it, so we don't need to check for
            # that here: only that the context is not None

            if not context["name"]:
                continue


            # go through the commands in this context, seeing if the
            # line matches one of them

            for cmd in self._parser[context["name"]]:
                match = cmd.ismatch(line_strip)

                if match:
                    if debug:
                        print("=> matches /^%s$/" % str(cmd.match))


                    # call the parse() method for this command

                    try:
                        new_config = cmd.parse(
                            context["config"], **match.groupdict())

                    except:
                        print("with: type(cmd)=" + repr(type(cmd)),
                              file=sys.stderr)

                        raise


                    # if this command specifies a new [sub] context is
                    # to be entered, change to that

                    if cmd.enter_context:
                        # if there are no commands registered for the
                        # new context, we can't enter it, so don't
                        # bother completing the new context information
                        #
                        # (the main loop assumes that the context
                        # exists, so this avoids that check by leaving
                        # the name None, from the default)

                        if cmd.enter_context not in self._parser:
                            if debug:
                                print("!> new context: %s unknown - ignoring"
                                          % cmd.enter_context)

                            break


                        if debug:
                            print("*> entering context:", cmd.enter_context)


                        new_context["name"] = cmd.enter_context


                        # use a more-specific configuration element
                        # dictionary, if one was returned by the
                        # parse() method

                        if new_config is not None:
                            new_context["config"] = new_config


                    # since we've found a matching command, don't bother
                    # looking for any more

                    break

                else:
                    if debug >= 2:
                        print("-> no match /^%s$/" % str(cmd.match))


            else:
                if debug:
                    print("=> no matches")


            # store the new context on the stack

            contexts.append(new_context)


        file.close()


        # perform any post-parsing processing

        self._post_parse_file()


    def _post_parse_file(self):
        """This method is called after parsing a file into the
        configuration dictionary with parse_file() (or as part of the
        constructor).

        In this abstract class, it does nothing, but can be overridden
        in child classes to perform any necessary processing.
        """

        pass



# ios.cmds
#
# Copyright (C) Robert Franklin <rcf34@cam.ac.uk>



"""Cisco IOS configuration commands module.

This module parses Cisco IOS configuration files into a dictionary.
"""



# --- imports ---



from deepops import deepsetdefault, deepget
from netaddr import IPNetwork

from net_contextdiff.config import IndentedContextualCommand

from .utils import (
    interface_canonicalize,
    ip_acl_ext_rule_canonicalize,
    ipv6_acl_rule_canonicalize,
    expand_set )



# cmds = []
#
# This is a list of classes, one for each IOS configuration mode
# command.  The command classes are defined at the global level and
# added to this list.
#
# The CiscoIOSConfig class adds these to the object upon instantiation,
# by the _add_commands() method.
#
# This was done to make it clearer how the commands are implemented.

cmds = []



# --- configuration command classes ---



# _Cmd is created to be a shorthand for the IndentedContextualCommand
# class as we'll be using it a lot

_Cmd = IndentedContextualCommand



# SYSTEM



class _Cmd_Comment(_Cmd):
    match = r"!.*"

cmds.append(_Cmd_Comment)


class _Cmd_Hostname(_Cmd):
    match = r"hostname (?P<hostname>\S+)"

    def parse(self, cfg, hostname):
        cfg["hostname"] = hostname

cmds.append(_Cmd_Hostname)



# INTERFACE ...



class _Cmd_Int(_Cmd):
    match = r"interface (?P<int_name>\S+)"
    enter_context = "interface"

    def parse(self, cfg, int_name):
        int_name = interface_canonicalize(int_name)

        i = deepsetdefault(cfg, "interface", int_name)

        # IOS has an odd behaviour that, when an interface is created in
        # configure mode, it will default to shutdown or not, depending
        # on its type; in startup configurations, however, they are
        # always not shutdown
        #
        # we default to not shutdown, unless this has been explicitly
        # overridden: this has the effect of 'no shutdown'ing the
        # interface, if it is being created
        i.setdefault("shutdown", False)

        return i

cmds.append(_Cmd_Int)


class _CmdContext_Int(_Cmd):
    context = "interface"


class _Cmd_Int_ARPTime(_CmdContext_Int):
    match = r"arp timeout (?P<time>\d+)"

    def parse(self, cfg, time):
        cfg["arp-timeout"] = int(time)

cmds.append(_Cmd_Int_ARPTime)


class _Cmd_Int_CDPEna(_CmdContext_Int):
    match = r"(?P<no>no )?cdp enable"

    def parse(self, cfg, no):
        # we allow CDP to be 'no cdp enable' to clear the CDP status
        cfg["cdp-enable"] = not no

cmds.append(_Cmd_Int_CDPEna)


class _Cmd_Int_ChnGrp(_CmdContext_Int):
    match = r"channel-group (?P<id_>\d+)(?P<mode> .+)?"

    def parse(self, cfg, id_, mode):
        cfg["channel-group"] = int(id_), mode

cmds.append(_Cmd_Int_ChnGrp)


class _Cmd_Int_Desc(_CmdContext_Int):
    match = r"description (?P<desc>.+)"

    def parse(self, cfg, desc):
        cfg["description"] = desc

cmds.append(_Cmd_Int_Desc)


class _Cmd_Int_Encap(_CmdContext_Int):
    match = r"encapsulation (?P<encap>dot1q \d+( native)?)"

    def parse(self, cfg, encap):
        # lower case the encapsulation definition as IOS stores 'dot1q'
        # as 'dot1Q'
        cfg["encapsulation"] = encap.lower()

cmds.append(_Cmd_Int_Encap)


class _Cmd_Int_IPAccGrp(_CmdContext_Int):
    match = r"ip access-group (?P<acl_name>\S+) (?P<dir_>in|out)"

    def parse(self, cfg, acl_name, dir_):
        cfg.setdefault("ip-access-group", {})[dir_] = acl_name

cmds.append(_Cmd_Int_IPAccGrp)


class _Cmd_Int_IPAddr(_CmdContext_Int):
    match = r"ip address (?P<addr>\S+ \S+)"

    def parse(self, cfg, addr):
        cfg["ip-address"] = addr

cmds.append(_Cmd_Int_IPAddr)


class _Cmd_Int_IPAddrSec(_CmdContext_Int):
    match = r"ip address (?P<addr>\S+ \S+) secondary"

    def parse(self, cfg, addr):
        # secondary address - record it in a list
        cfg.setdefault("ip-address-secondary", set()).add(addr)

cmds.append(_Cmd_Int_IPAddrSec)


class _Cmd_Int_IPFlowMon(_CmdContext_Int):
    match = r"ip flow monitor (?P<flowmon>\S+) (?P<dir_>input|output)"

    def parse(self, cfg, flowmon, dir_):
        deepsetdefault(cfg, "ip-flow-monitor")[dir_] = flowmon

cmds.append(_Cmd_Int_IPFlowMon)


class _Cmd_Int_IPHlprAddr(_CmdContext_Int):
    match = r"ip helper-address (?P<addr>(global )?\S+)"

    def parse(self, cfg, addr):
        cfg.setdefault("ip-helper-address", set()).add(addr)

cmds.append(_Cmd_Int_IPHlprAddr)


class _Cmd_Int_IPIGMPVer(_CmdContext_Int):
    match = r"ip igmp version (?P<ver>\S+)"

    def parse(self, cfg, ver):
        cfg["ip-igmp-version"] = ver

cmds.append(_Cmd_Int_IPIGMPVer)


class _Cmd_Int_IPMcastBdry(_CmdContext_Int):
    match = r"ip multicast boundary (?P<acl>\S+)"

    def parse(self, cfg, acl):
        cfg["ip-multicast-boundary"] = acl

cmds.append(_Cmd_Int_IPMcastBdry)


class _Cmd_Int_IPPIMMode(_CmdContext_Int):
    match = r"ip pim (?P<mode>(sparse|dense|sparse-dense)-mode)"

    def parse(self, cfg, mode):
        cfg.setdefault("ip-pim", {})["mode"] = mode

cmds.append(_Cmd_Int_IPPIMMode)


class _Cmd_Int_IPPIMBSRBdr(_CmdContext_Int):
    match = r"ip pim bsr-border"

    def parse(self, cfg):
        cfg.setdefault("ip-pim", {})["bsr-border"] = True

cmds.append(_Cmd_Int_IPPIMBSRBdr)


class _Cmd_Int_IPProxyARP(_CmdContext_Int):
    match = r"(?P<no>no )?ip proxy-arp"

    def parse(self, cfg, no):
        cfg["ip-proxy-arp"] = not no

cmds.append(_Cmd_Int_IPProxyARP)


class _Cmd_Int_IPVerifyUni(_CmdContext_Int):
    match = r"ip verify unicast (?P<opt>.+)"

    def parse(self, cfg, opt):
        cfg["ip-verify-unicast"] = opt

cmds.append(_Cmd_Int_IPVerifyUni)


class _Cmd_Int_IPv6Addr(_CmdContext_Int):
    match = r"ipv6 address (?P<addr>\S+)"

    def parse(self, cfg, addr):
        # IPv6 addresses involve letters so we lower case for
        # consistency
        cfg.setdefault("ipv6-address", set()).add(addr.lower())

cmds.append(_Cmd_Int_IPv6Addr)


class _Cmd_Int_IPv6MultBdry(_CmdContext_Int):
    match = r"ipv6 multicast boundary scope (?P<scope>\S+)"

    def parse(self, cfg, scope):
        cfg["ipv6-multicast-boundary-scope"] = scope

cmds.append(_Cmd_Int_IPv6MultBdry)


class _Cmd_Int_IPv6PIMBSRBdr(_CmdContext_Int):
    match = r"ipv6 pim bsr border"

    def parse(self, cfg):
        cfg.setdefault("ipv6-pim", {})["bsr-border"] = True

cmds.append(_Cmd_Int_IPv6PIMBSRBdr)


class _Cmd_Int_IPv6TrafFilt(_CmdContext_Int):
    match = r"ipv6 traffic-filter (?P<acl_name>\S+) (?P<dir_>in|out)"

    def parse(self, cfg, acl_name, dir_):
        cfg.setdefault("ipv6-traffic-filter", {})[dir_] = acl_name

cmds.append(_Cmd_Int_IPv6TrafFilt)


class _Cmd_Int_IPv6VerifyUni(_CmdContext_Int):
    match = r"ipv6 verify unicast (?P<opt>.+)"

    def parse(self, cfg, opt):
        cfg["ipv6-verify-unicast"] = opt

cmds.append(_Cmd_Int_IPv6VerifyUni)


class _Cmd_Int_ServPol(_CmdContext_Int):
    match = r"service-policy (?P<policy>.+)"

    def parse(self, cfg, policy):
        cfg.setdefault("service-policy", set()).add(policy)

cmds.append(_Cmd_Int_ServPol)


class _Cmd_Int_Shutdown(_CmdContext_Int):
    match = r"(?P<no>no )?shutdown"

    def parse(self, cfg, no):
        cfg["shutdown"] = not no

cmds.append(_Cmd_Int_Shutdown)


class _Cmd_Int_StandbyIP(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) ip (?P<addr>\S+)"

    def parse(self, cfg, grp, addr):
        deepsetdefault(
            cfg, "standby", "group", int(grp))["ip"] = addr

cmds.append(_Cmd_Int_StandbyIP)


class _Cmd_Int_StandbyIPSec(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) ip (?P<addr>\S+) secondary"

    def parse(self, cfg, grp, addr):
        deepsetdefault(
            cfg, "standby", "group", int(grp), "ip-secondary",
            last=set()).add(addr)

cmds.append(_Cmd_Int_StandbyIPSec)


class _Cmd_Int_StandbyIPv6(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) ipv6 (?P<addr>\S+)"

    def parse(self, cfg, grp, addr):
        deepsetdefault(
            cfg, "standby", "group", int(grp), "ipv6", last=set()).add(addr)

cmds.append(_Cmd_Int_StandbyIPv6)


class _Cmd_Int_StandbyPreempt(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) preempt"

    def parse(self, cfg, grp):
        deepsetdefault(
            cfg, "standby", "group", int(grp))["preempt"] = True

cmds.append(_Cmd_Int_StandbyPreempt)


class _Cmd_Int_StandbyPri(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) priority (?P<pri>\d+)"

    def parse(self, cfg, grp, pri):
        deepsetdefault(
            cfg, "standby", "group", int(grp))["priority"] = int(pri)

cmds.append(_Cmd_Int_StandbyPri)


class _Cmd_Int_StandbyTimers(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) timers (?P<timers>\d+ \d+)"

    def parse(self, cfg, grp, timers):
        deepsetdefault(
            cfg, "standby", "group", int(grp))["timers"] = timers

cmds.append(_Cmd_Int_StandbyTimers)


class _Cmd_Int_StandbyTrk(_CmdContext_Int):
    match = r"standby (?P<grp>\d+) track (?P<obj>\d+)( (?P<extra>.+))?"

    def parse(self, cfg, grp, obj, extra):
        deepsetdefault(
            cfg, "standby", "group", int(grp), "track")[obj] = extra

cmds.append(_Cmd_Int_StandbyTrk)


class _Cmd_Int_StandbyVer(_CmdContext_Int):
    match = r"standby version (?P<ver>\d)"

    def parse(self, cfg, ver):
        deepsetdefault(cfg, "standby")["version"] = int(ver)

cmds.append(_Cmd_Int_StandbyVer)


class _Cmd_Int_StormCtrl(_CmdContext_Int):
    match = r"storm-control (?P<traffic>\S+) level (?P<level>[0-9.]+)"

    def parse(self, cfg, traffic, level):
        deepsetdefault(cfg, "storm-control")[traffic] = float(level)

cmds.append(_Cmd_Int_StormCtrl)


class _Cmd_Int_SwPort(_CmdContext_Int):
    match = r"(?P<no>no )?switchport"

    def parse(self, cfg, no):
        cfg["switchport"] = not no

cmds.append(_Cmd_Int_SwPort)


class _Cmd_Int_SwPortMode(_CmdContext_Int):
    match = r"switchport mode (?P<mode>\S+)"

    def parse(self, cfg, mode):
        cfg["switchport-mode"] = mode

cmds.append(_Cmd_Int_SwPortMode)


class _Cmd_Int_SwPortNoNeg(_CmdContext_Int):
    match = r"switchport nonegotiate"

    def parse(self, cfg):
        cfg["switchport-nonegotiate"] = True

cmds.append(_Cmd_Int_SwPortNoNeg)


class _Cmd_Int_SwPortTrkNtv(_CmdContext_Int):
    match = r"switchport trunk native vlan (?P<vlan>\d+)"

    def parse(self, cfg, vlan):
        cfg["switchport-trunk-native"] = int(vlan)

cmds.append(_Cmd_Int_SwPortTrkNtv)


class _Cmd_Int_SwPortTrkAlw(_CmdContext_Int):
    match = r"switchport trunk allowed vlan (add )?(?P<vlans>[0-9,-]+)"

    def parse(self, cfg, vlans):
        cfg.setdefault("switchport-trunk-allow", set()).update(
            expand_set(vlans))

cmds.append(_Cmd_Int_SwPortTrkAlw)


class _Cmd_Int_VRFFwd(_CmdContext_Int):
    match = (r"vrf forwarding (?P<name>\S+)")

    def parse(self, cfg, name):
        cfg["vrf-forwarding"] = name

cmds.append(_Cmd_Int_VRFFwd)


class _Cmd_Int_XConn(_CmdContext_Int):
    match = r"xconnect (?P<remote>[0-9.]+ \d+ .+)"

    def parse(self, cfg, remote):
        cfg["xconnect"] = remote

cmds.append(_Cmd_Int_XConn)



# IP ACCESS-LIST STANDARD



class _Cmd_ACLStdRule(_Cmd):
    match = r"access-list (?P<num>\d{1,2}|1[3-9]\d{2}) (?P<rule>.+)"

    def parse(self, cfg, num, rule):
        deepsetdefault(
            cfg, "ip-access-list-standard", num, last=[]).append(rule)

cmds.append(_Cmd_ACLStdRule)


class _Cmd_IPACL_Std(_Cmd):
    match = r"ip access-list standard (?P<acl_name>.+)"
    enter_context = "ip-acl_std"

    def parse(self, cfg, acl_name):
        return deepsetdefault(
                   cfg, "ip-access-list-standard", acl_name, last=[])

cmds.append(_Cmd_IPACL_Std)


class _Cmd_IPACL_Std_Rule(_Cmd):
    context = "ip-acl_std"
    match = r"(?P<rule>(permit|deny) +.+)"

    def parse(self, cfg, rule):
        cfg.append(rule)

cmds.append(_Cmd_IPACL_Std_Rule)


class _Cmd_ACLExtRule(_Cmd):
    match = r"access-list (?P<num>1\d{2}|2[0-6]\d{2}) (?P<rule>.+)"

    def parse(self, cfg, num, rule):
        deepsetdefault(
            cfg, "ip-access-list-extended", num, last=[]
            ).append(ip_acl_ext_rule_canonicalize(rule))

cmds.append(_Cmd_ACLExtRule)


class _Cmd_IPACL_Ext(_Cmd):
    match = r"ip access-list extended (?P<name>.+)"
    enter_context = "ip-acl_ext"

    def parse(self, cfg, name):
        return deepsetdefault(cfg, "ip-access-list-extended", name, last=[])

cmds.append(_Cmd_IPACL_Ext)


class _Cmd_IPACL_Ext_Rule(_Cmd):
    context = "ip-acl_ext"
    match = r"(?P<rule>(permit|deny) +.+)"

    def parse(self, cfg, rule):
        cfg.append(ip_acl_ext_rule_canonicalize(rule))

cmds.append(_Cmd_IPACL_Ext_Rule)



# IPV6 ACCESS-LIST ...



class _Cmd_IPv6ACL(_Cmd):
    match = r"ipv6 access-list (?P<name>.+)"
    enter_context = "ipv6-acl"

    def parse(self, cfg, name):
        return deepsetdefault(cfg, "ipv6-access-list", name, last=[])

cmds.append(_Cmd_IPv6ACL)


class _Cmd_IPv6ACL_Rule(_Cmd):
    context = "ipv6-acl"
    match = r"(?P<rule>(permit|deny) +.+)"

    def parse(self, cfg, rule):
        cfg.append(ipv6_acl_rule_canonicalize(rule))

cmds.append(_Cmd_IPv6ACL_Rule)



# IP[V6] PREFIX-LIST ...



class _Cmd_IPPfx(_Cmd):
    match = r"ip prefix-list (?P<list_>\S+) (seq \d+ )?(?P<rule>.+)"

    def parse(self, cfg, list_, rule):
        deepsetdefault(cfg, "ip-prefix-list", list_, last=[]).append(rule)

cmds.append(_Cmd_IPPfx)


class _Cmd_IPv6Pfx(_Cmd):
    match = r"ipv6 prefix-list (?P<list_>\S+) (seq \d+ )?(?P<rule>.+)"

    def parse(self, cfg, list_, rule):
        deepsetdefault(
            cfg, "ipv6-prefix-list", list_, last=[]).append(rule.lower())

cmds.append(_Cmd_IPv6Pfx)



# IP[V6] ROUTE ...



class _Cmd_IPRoute(_Cmd):
    match = r"ip route (?P<route>.+)"

    def parse(self, cfg, route):
        cfg.setdefault("ip-route", set()).add(route)

cmds.append(_Cmd_IPRoute)


class _Cmd_IPv6Route(_Cmd):
    match = r"ipv6 route (?P<route>.+)"

    def parse(self, cfg, route):
        # IPv6 addresses involve letters so we lower case for
        # consistency
        cfg.setdefault("ipv6-route", set()).add(route.lower())

cmds.append(_Cmd_IPv6Route)



# [NO] SPANNING-TREE ...



class _Cmd_NoSTP(_Cmd):
    match = r"no spanning-tree vlan (?P<tags>[-0-9,]+)"

    def parse(self, cfg, tags):
        cfg.setdefault(
            "no-spanning-tree-vlan", set()).update(expand_set(tags))

cmds.append(_Cmd_NoSTP)


class _Cmd_STPPri(_Cmd):
    match = r"spanning-tree vlan (?P<tags>[-0-9,]+) priority (?P<pri>\d+)"

    def parse(self, cfg, tags, pri):
        cfg_stp_pri = cfg.setdefault("spanning-tree-vlan-priority", {})
        for tag in expand_set(tags):
            cfg_stp_pri[int(tag)] = int(pri)

cmds.append(_Cmd_STPPri)



# TRACK ...



class _Cmd_Track(_Cmd):
    match = r"track (?P<obj>\d+)"
    enter_context = "track"

    def parse(self, cfg, obj):
        # if there is no criterion, we're modifying an existing object,
        # which must have already been defined, so we deliberately don't
        # create it with deepsetdefault() but just deepget() it with
        # default_error set, to force an error here, if it doesn't exist
        return deepget(cfg, "track", int(obj), default_error=True)

cmds.append(_Cmd_Track)


class _CmdContext_Track(_Cmd):
    context = "track"


class _Cmd_Track_Delay(_CmdContext_Track):
    match = r"delay (?P<delay>.+)"

    def parse(self, cfg, delay):
        cfg["delay"] = delay

cmds.append(_Cmd_Track_Delay)


class _Cmd_Track_IPVRF(_CmdContext_Track):
    match = r"ip vrf (?P<vrf_name>\S+)"

    def parse(self, cfg, vrf_name):
        cfg["ip-vrf"] = vrf_name

cmds.append(_Cmd_Track_IPVRF)


class _Cmd_Track_IPv6VRF(_CmdContext_Track):
    match = r"ipv6 vrf (?P<vrf_name>\S+)"

    def parse(self, cfg, vrf_name):
        cfg["ipv6-vrf"] = vrf_name

cmds.append(_Cmd_Track_IPv6VRF)


class _Cmd_Track_Obj(_CmdContext_Track):
    match = r"object (?P<obj>.+)"

    def parse(self, cfg, obj):
        deepsetdefault(cfg, "object", last=set()).add(obj)

cmds.append(_Cmd_Track_Obj)


class _Cmd_TrackRoute(_Cmd):
    match = (r"track (?P<obj>\d+)"
           r" (?P<proto>ip|ipv6) route"
           r" (?P<net>[0-9a-fA-F.:]+/\d+|[0-9.]+ [0-9.]+)"
           r" (?P<extra>metric .+|reachability)")
    enter_context = "track"

    def parse(self, cfg, obj, proto, net, extra):
        # the 'net' can be in 'network netmask' or CIDR format, but the
        # netaddr.IPNetwork() object requires a slash between the
        # network and netmask, so we just change the space to a slash
        net = IPNetwork(net.replace(" ", "/"))

        # reconstruct a normalised version of the criterion
        criterion = ("%s route %s %s" % (proto, net, extra))

        # create the new track object and store the criterion in it
        t = deepsetdefault(cfg, "track", int(obj))
        t["criterion"] = criterion

        # return the track object for the new context
        return t

cmds.append(_Cmd_TrackRoute)


class _Cmd_TrackOther(_Cmd):
    match = r"track (?P<obj>\d+) (?P<other>(interface .+|list .+|stub-object))"
    enter_context = "track"

    def parse(self, cfg, obj, other):
        t = deepsetdefault(cfg, "track", int(obj))
        t["criterion"] = other

        return t

cmds.append(_Cmd_TrackOther)



# VLAN ...



class _Cmd_VLAN(_Cmd):
    match = r"vlan (?P<tag>\d+)"
    enter_context = "vlan"

    def parse(self, cfg, tag):
        # create the VLAN configuration entry, setting an 'exists' key
        # as we might stop other information in here that isn't in the
        # VLAN definition itself in IOS (e.g. STP priority) in future
        v = deepsetdefault(cfg, "vlan", int(tag))
        v["exists"] = True

        return v

cmds.append(_Cmd_VLAN)


class _CmdContext_VLAN(_Cmd):
    context = "vlan"


class _Cmd_VLAN_Name(_CmdContext_VLAN):
    match = r"name (?P<name>\S+)"

    def parse(self, cfg, name):
        cfg["name"] = name

cmds.append(_Cmd_VLAN_Name)
